"""
Test the core grains
"""

import pytest

import salt.loader
import salt.utils.platform
from tests.support.case import ModuleCase
from tests.support.mixins import LoaderModuleMockMixin

if salt.utils.platform.is_windows():
    try:
        import salt.modules.reg
    except ImportError:
        pass


@pytest.mark.windows_whitelisted
class TestGrainsCore(ModuleCase):
    """
    Test the core grains grains
    """

    @pytest.mark.skip_unless_on_platforms(freebsd=True, openbsd=True)
    def test_freebsd_openbsd_mem_total(self):
        """
        test grains['mem_total']
        """
        physmem = self.run_function("sysctl.get", ["hw.physmem"])
        self.assertEqual(
            self.run_function("grains.items")["mem_total"], int(physmem) // 1048576
        )

    @pytest.mark.skip_unless_on_openbsd
    def test_openbsd_swap_total(self):
        """
        test grains['swap_total']
        """
        swapmem = self.run_function("cmd.run", ["swapctl -sk"]).split(" ")[1]
        self.assertEqual(
            self.run_function("grains.items")["swap_total"], int(swapmem) // 1048576
        )


@pytest.mark.windows_whitelisted
class TestGrainsReg(ModuleCase, LoaderModuleMockMixin):
    """
    Test the core windows grains
    """

    def setup_loader_modules(self):
        self.opts = opts = salt.config.DEFAULT_MINION_OPTS.copy()
        utils = salt.loader.utils(opts, whitelist=["reg"])
        return {salt.modules.reg: {"__opts__": opts, "__utils__": utils}}

    @pytest.mark.skip_unless_on_windows
    @pytest.mark.slow_test
    def test_win_cpu_model(self):
        """
        test grains['cpu_model']
        """
        cpu_model_text = salt.modules.reg.read_value(
            "HKEY_LOCAL_MACHINE",
            "HARDWARE\\DESCRIPTION\\System\\CentralProcessor\\0",
            "ProcessorNameString",
        ).get("vdata")
        self.assertEqual(self.run_function("grains.items")["cpu_model"], cpu_model_text)



"""
Functions for adding yaml encoding to the jinja context
"""

import io
import sys

import yaml  # pylint: disable=blacklisted-import

from salt.utils.decorators.jinja import jinja_filter


@jinja_filter()
def yaml_dquote(text):
    """
    Make text into a double-quoted YAML string with correct escaping
    for special characters.  Includes the opening and closing double
    quote characters.
    """
    with io.StringIO() as ostream:
        yemitter = yaml.emitter.Emitter(ostream, width=sys.maxsize)
        yemitter.write_double_quoted(str(text))
        return ostream.getvalue()


@jinja_filter()
def yaml_squote(text):
    """
    Make text into a single-quoted YAML string with correct escaping
    for special characters.  Includes the opening and closing single
    quote characters.
    """
    with io.StringIO() as ostream:
        yemitter = yaml.emitter.Emitter(ostream, width=sys.maxsize)
        yemitter.write_single_quoted(str(text))
        return ostream.getvalue()


@jinja_filter()
def yaml_encode(data):
    """
    A simple YAML encode that can take a single-element datatype and return
    a string representation.
    """
    yrepr = yaml.representer.SafeRepresenter()
    ynode = yrepr.represent_data(data)
    if not isinstance(ynode, yaml.ScalarNode):
        raise TypeError(
            "yaml_encode() only works with YAML scalar data; failed for {}".format(
                type(data)
            )
        )

    tag = ynode.tag.rsplit(":", 1)[-1]
    ret = ynode.value

    if tag == "str":
        ret = yaml_dquote(ynode.value)

    return ret



"""
Manage the context a module loaded by Salt's loader
"""

import collections.abc
import contextlib
import copy

try:
    # Try the stdlib C extension first
    import _contextvars as contextvars
except ImportError:
    # Py<3.7
    import contextvars

import salt.exceptions

DEFAULT_CTX_VAR = "loader_ctxvar"

loader_ctxvar = contextvars.ContextVar(DEFAULT_CTX_VAR)


@contextlib.contextmanager
def loader_context(loader):
    """
    A context manager that sets and un-sets the loader context
    """
    tok = loader_ctxvar.set(loader)
    try:
        yield
    finally:
        loader_ctxvar.reset(tok)


class NamedLoaderContext(collections.abc.MutableMapping):
    """
    A NamedLoaderContext object is injected by the loader providing access to
    Salt's 'magic dunders' (__salt__, __utils__, etc).
    """

    def __init__(self, name, loader_context, default=None):
        self.name = name
        self.loader_context = loader_context
        self.default = default

    def with_default(self, default):
        return NamedLoaderContext(self.name, self.loader_context, default=default)

    def loader(self):
        """
        The LazyLoader in the current context. This will return None if there
        is no context established.
        """
        try:
            return self.loader_context.loader()
        except AttributeError:
            return None

    def eldest_loader(self):
        if self.loader() is None:
            return None
        loader = self.loader()
        while loader.parent_loader is not None:
            loader = loader.parent_loader
        return loader

    def value(self):
        """
        The value of the current for this context
        """
        loader = self.loader()
        if loader is None:
            return self.default
        if self.name == loader.pack_self:
            return loader
        elif self.name == "__context__":
            return loader.pack[self.name]
        elif self.name == "__opts__":
            return loader.pack[self.name]
        try:
            return loader.pack[self.name]
        except KeyError:
            raise salt.exceptions.LoaderError(
                f"LazyLoader does not have a packed value for: {self.name}"
            )

    def get(self, key, default=None):
        return self.value().get(key, default)

    def __getitem__(self, item):
        return self.value()[item]

    def __contains__(self, item):
        return item in self.value()

    def __setitem__(self, item, value):
        self.value()[item] = value

    def __bool__(self):
        return bool(self.value())

    def __len__(self):
        return self.value().__len__()

    def __iter__(self):
        return self.value().__iter__()

    def __delitem__(self, item):
        return self.value().__delitem__(item)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.loader_context == other.loader_context and self.name == other.name

    def __getstate__(self):
        return {
            "name": self.name,
            "loader_context": self.loader_context,
            "default": None,
        }

    def __setstate__(self, state):
        self.name = state["name"]
        self.loader_context = state["loader_context"]
        self.default = state["default"]

    def __getattr__(self, name):
        return getattr(self.value(), name)

    def __deepcopy__(self, memo):
        default = copy.deepcopy(self.default)
        return self.__class__(self.name, self.loader_context, default)

    def missing_fun_string(self, name):
        return self.loader().missing_fun_string(name)


class LoaderContext:
    """
    A loader context object, this object is injected at <loaded
    module>.__salt_loader__ by the Salt loader. It is responsible for providing
    access to the current context's loader
    """

    def __init__(self, loader_ctxvar=loader_ctxvar):
        self.loader_ctxvar = loader_ctxvar

    def __getitem__(self, item):
        return self.loader[item]

    def loader(self):
        """
        Return the LazyLoader in the current context. If there is no value set raise an AttributeError
        """
        try:
            return self.loader_ctxvar.get()
        except LookupError:
            raise AttributeError("No loader context")

    def named_context(self, name, default=None, ctx_class=NamedLoaderContext):
        """
        Return a NamedLoaderContext instance which will use this LoaderContext
        """
        return ctx_class(name, self, default)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.loader_ctxvar == other.loader_ctxvar

    def __getstate__(self):
        return {"varname": self.loader_ctxvar.name}

    def __setstate__(self, state):
        self.loader_ctxvar = contextvars.ContextVar(state["varname"])



import pytest

import salt.cloud.clouds.openstack as openstack
from salt.utils import dictupdate
from tests.support.mock import MagicMock, call, patch

# pylint: disable=confusing-with-statement


class MockImage:
    name = "image name"
    id = "image id"


class MockNode:
    name = "node name"
    id = "node id"
    flavor = MockImage()
    status = "node status"

    def __init__(self, image):
        self.image = image

    def __iter__(self):
        return iter(())


class MockConn:
    def __init__(self, image):
        self.node = MockNode(image)

    def get_image(self, *args, **kwargs):
        return self.node.image

    def get_flavor(self, *args, **kwargs):
        return self.node.flavor

    def get_server(self, *args, **kwargs):
        return self.node

    def list_servers(self, *args, **kwargs):
        return [self.node]


@pytest.fixture
def configure_loader_modules():
    return {
        openstack: {
            "__active_provider_name__": "",
            "__opts__": {
                "providers": {
                    "my-openstack-cloud": {
                        "openstack": {
                            "auth": "daenerys",
                            "region_name": "westeros",
                            "cloud": "openstack",
                        }
                    }
                }
            },
        }
    }


@pytest.fixture
def expected_config_stuff():
    vm = {"asdf": ...}

    with patch("copy.deepcopy", autospec=True, return_value=42), patch.dict(
        openstack.__opts__, {"foo": "bar"}
    ):
        yield vm


def test_when_getting_cloud_config_values_expected_args_should_be_provided():
    expected_vm = "whatever"
    expected_calls = call(
        "ignore_cidr", expected_vm, openstack.__opts__, default="", search_global=False
    )


@pytest.mark.parametrize(
    "comment,example_ip,ignored_cidr,expected",
    [
        ("ip is in ignore_cidr string", "203.0.113.1", "203.0.113.0/24", True),
        ("ip is not in ignore_cidr string", "192.0.2.1", "203.0.113.0/24", False),
        ("ignore_cidr is empty", "192.0.2.1", "", False),
        ("ignore_cidr is False", "192.0.2.1", False, False),
        ("ignore_cidr is None", "192.0.2.1", None, False),
        (
            "ip is in ignore_cidr list",
            "192.0.2.1",
            ["192.0.2.0/24", "203.0.113.0/24"],
            True,
        ),
        (
            "ip is not in ignore_cidr list",
            "192.0.2.1",
            ["198.51.100.0/24", "203.0.113.0/24"],
            False,
        ),
    ],
)
def test_when_ignore_cidr_is_configured_and_ip_is_provided_result_is_expected(
    comment, example_ip, ignored_cidr, expected
):
    with patch(
        "salt.config.get_cloud_config_value", autospec=True, return_value=ignored_cidr
    ):
        result = openstack.ignore_cidr("fnord", example_ip)

    assert result is expected


@pytest.mark.parametrize(
    "comment,example_ips,ignored_cidr,expected",
    [
        (
            "ignore_cidr matches first 2 ips, expected value will be first ip that"
            " doesn't match cidr.",
            ["203.0.113.1", "203.0.113.2", "192.0.2.1", "192.0.2.2"],
            "203.0.113.0/24",
            "192.0.2.1",
        ),
        (
            "ignore_cidr matches 2nd 2 IPs, expected value will be first ip in list. ",
            ["203.0.113.1", "203.0.113.2", "192.0.2.1", "192.0.2.2"],
            "192.0.2.0/24",
            "203.0.113.1",
        ),
        (
            "ignore_cidr doesn't match any IPs, expected value will be first ip in"
            " list.",
            ["203.0.113.1", "203.0.113.2", "192.0.2.1", "192.0.2.2"],
            "198.51.100.0/24",
            "203.0.113.1",
        ),
        (
            "ignore_cidr matches all IPs, expected value will be False.",
            ["203.0.113.1", "203.0.113.2", "203.0.113.3", "203.0.113.4"],
            "203.0.113.0/24",
            False,
        ),
        (
            "When ignore_cidr is not set, return first ip",
            ["203.0.113.1", "203.0.113.2", "192.0.2.1", "192.0.2.2"],
            None,
            "203.0.113.1",
        ),
    ],
)
def test_preferred_ip_function_returns_expected(
    comment, example_ips, ignored_cidr, expected
):
    with patch(
        "salt.config.get_cloud_config_value", autospec=True, return_value=ignored_cidr
    ):
        result = openstack.preferred_ip("fnord", example_ips)

    assert result is expected


def test_get_configured_provider_bad():
    with patch.dict(openstack.__opts__, {"providers": {}}):
        result = openstack.get_configured_provider()
        assert result is False


def test_get_configured_provider_auth():
    config = {
        "region_name": "westeros",
        "auth": "daenerys",
    }
    with patch.dict(
        openstack.__opts__,
        {"providers": {"my-openstack-cloud": {"openstack": config}}},
    ):
        result = openstack.get_configured_provider()
        assert config == result


def test_get_configured_provider_cloud():
    config = {
        "region_name": "westeros",
        "cloud": "foo",
    }
    with patch.dict(
        openstack.__opts__,
        {"providers": {"my-openstack-cloud": {"openstack": config}}},
    ):
        result = openstack.get_configured_provider()
        assert config == result


def test_get_dependencies():
    HAS_SHADE = (True, "Please install newer version of shade: >= 1.19.0")
    with patch("salt.cloud.clouds.openstack.HAS_SHADE", HAS_SHADE):
        result = openstack.get_dependencies()
        assert result is True


def test_get_dependencies_no_shade():
    HAS_SHADE = (False, "Install pypi module shade >= 1.19.0")
    with patch("salt.cloud.clouds.openstack.HAS_SHADE", HAS_SHADE):
        result = openstack.get_dependencies()
        assert result is False


def test_list_nodes_full_image_str():
    node_image = "node image"
    conn = MockConn(node_image)
    with patch("salt.cloud.clouds.openstack._get_ips", return_value=[]):
        ret = openstack.list_nodes_full(conn=conn)
        assert ret[conn.node.name]["image"] == node_image


def test_list_nodes_full_image_obj():
    conn = MockConn(MockImage())
    with patch("salt.cloud.clouds.openstack._get_ips", return_value=[]):
        ret = openstack.list_nodes_full(conn=conn)
        assert ret[conn.node.name]["image"] == MockImage.name


def test_show_instance():
    conn = MockConn(MockImage())
    with patch("salt.cloud.clouds.openstack._get_ips", return_value=[]):
        ret = openstack.show_instance(conn.node.name, conn=conn, call="action")
        assert ret["image"] == MockImage.name


def test_request_instance_should_use_provided_connection_if_not_None():
    fake_conn = MagicMock()

    patch_get_conn = patch("salt.cloud.clouds.openstack.get_conn", autospec=True)
    patch_utils = patch.dict(
        openstack.__utils__,
        {"cloud.check_name": MagicMock(), "dictupdate.update": dictupdate.update},
    )
    patch_shade = patch.object(
        openstack, "shade.exc.OpenStackCloudException", Exception, create=True
    )

    with patch_get_conn as fake_get_conn, patch_utils, patch_shade:
        openstack.request_instance(
            vm_={"name": "fnord", "driver": "fnord"}, conn=fake_conn
        )

        fake_get_conn.assert_not_called()


def test_request_instance_should_create_conn_if_provided_is_None():
    none_conn = None

    patch_get_conn = patch("salt.cloud.clouds.openstack.get_conn", autospec=True)
    patch_utils = patch.dict(
        openstack.__utils__,
        {"cloud.check_name": MagicMock(), "dictupdate.update": dictupdate.update},
    )
    patch_shade = patch.object(
        openstack, "shade.exc.OpenStackCloudException", Exception, create=True
    )

    with patch_get_conn as fake_get_conn, patch_utils, patch_shade:
        openstack.request_instance(
            vm_={"name": "fnord", "driver": "fnord"}, conn=none_conn
        )

        fake_get_conn.assert_called_once_with()


# According to
# https://docs.openstack.org/shade/latest/user/usage.html#shade.OpenStackCloud.create_server
# the `network` parameter can be:
# (optional) Network dict or name or ID to attach the server to.
# Mutually exclusive with the nics parameter. Can also be be a list of
# network names or IDs or network dicts.
#
# Here we're testing a normal dictionary
def test_request_instance_should_be_able_to_provide_a_dictionary_for_network():
    fake_conn = MagicMock()
    expected_network = {"foo": "bar"}
    vm_ = {"name": "fnord", "driver": "fnord", "network": expected_network}
    patch_utils = patch.dict(
        openstack.__utils__,
        {"cloud.check_name": MagicMock(), "dictupdate.update": dictupdate.update},
    )
    with patch_utils:
        openstack.request_instance(vm_=vm_, conn=fake_conn)

        call_kwargs = fake_conn.create_server.mock_calls[0][-1]
        assert call_kwargs["network"] == expected_network


# Here we're testing the list of dictionaries
def test_request_instance_should_be_able_to_provide_a_list_of_dictionaries_for_network():
    fake_conn = MagicMock()
    expected_network = [{"foo": "bar"}, {"bang": "quux"}]
    vm_ = {"name": "fnord", "driver": "fnord", "network": expected_network}
    patch_utils = patch.dict(
        openstack.__utils__,
        {"cloud.check_name": MagicMock(), "dictupdate.update": dictupdate.update},
    )
    with patch_utils:
        openstack.request_instance(vm_=vm_, conn=fake_conn)

        call_kwargs = fake_conn.create_server.mock_calls[0][-1]
        assert call_kwargs["network"] == expected_network


# Here we're testing for names/IDs
def test_request_instance_should_be_able_to_provide_a_list_of_single_ids_or_names_for_network():
    fake_conn = MagicMock()
    expected_network = ["foo", "bar", "bang", "fnord1", "fnord2"]
    vm_ = {"name": "fnord", "driver": "fnord", "network": expected_network}
    patch_utils = patch.dict(
        openstack.__utils__,
        {"cloud.check_name": MagicMock(), "dictupdate.update": dictupdate.update},
    )
    with patch_utils:
        openstack.request_instance(vm_=vm_, conn=fake_conn)

        call_kwargs = fake_conn.create_server.mock_calls[0][-1]
        assert call_kwargs["network"] == expected_network


# Testing that we get a dict that we expect for create_server
def test__clean_create_kwargs():
    params = {
        "name": "elmer",
        "image": "mirrormirror",
        "flavor": "chocolate",
        "auto_ip": True,
        "ips": ["hihicats"],
        "ip_pool": "olympic",
        "root_volume": "iamgroot",
        "boot_volume": "pussnboots",
        "terminate_volume": False,
        "volumes": ["lots", "of", "books"],
        "meta": {"full": "meta"},
        "files": {"shred": "this"},
        "reservation_id": "licenseandregistration",
        "security_groups": ["wanna", "play", "repeat"],
        "key_name": "clortho",
        "availability_zone": "callmemaybe",
        "block_device_mapping": [{"listof": "dicts"}],
        "block_device_mapping_v2": [{"listof": "dicts"}],
        "nics": ["thats", "me"],
        "scheduler_hints": {"so": "many"},
        "config_drive": True,
        "disk_config": "donkey",
        "admin_pass": "password",
        "wait": False,
        "timeout": 30,
        "reuse_ips": True,
        "network": ["also", "a", "dict"],
        "boot_from_volume": True,
        "volume_size": 30,
        "nat_destination": "albuquerque",
        "group": "ledzeppelin",
        "userdata": "needmoreinput",
        "thisgetsdropped": "yup",
    }
    patch_utils = patch.dict(
        openstack.__utils__,
        {"dictupdate.update": dictupdate.update},
    )
    with patch_utils:
        ret = openstack._clean_create_kwargs(**params)
        params.pop("thisgetsdropped")
        assert params == ret



"""
    :codeauthor: Gareth J. Greenaway <gareth@saltstack.com>
"""

import pytest

import salt.modules.cmdmod as cmdmod
import salt.modules.win_system as win_system
from tests.support.mock import MagicMock, patch

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
]


@pytest.fixture
def configure_loader_modules():
    return {win_system: {"__salt__": {"cmd.run": cmdmod.run}}}


def test_lock():
    """
    Test locking workstation
    """
    with patch("ctypes.windll.user32.LockWorkStation", MagicMock()):
        ret = win_system.lock()
        assert ret


def test_get_computer_name():
    """
    Test getting the computer name
    """
    ret = win_system.get_computer_name()
    assert ret


def test_set_computer_name():
    """
    Test getting the computer name
    """
    with patch("ctypes.windll.kernel32.SetComputerNameExW", MagicMock()):
        ret = win_system.set_computer_name("test_hostname")
        assert ret


def test_get_hostname():
    """
    Test getting the computer name
    """
    ret = win_system.get_hostname()
    assert ret


def test_set_system_time():
    """
    Test setting the system time
    """
    with patch.object(win_system, "set_system_date_time", MagicMock()):
        ret = win_system.set_system_time("12:05 AM")
        assert ret


def test_get_system_time():
    """
    Test setting the system time
    """
    ret = win_system.get_system_time()
    assert ret


def test_set_system_date_time():
    """
    Test setting the system time
    """
    get_local_time = (2022, 6, 3, 11, 0, 0, 0)
    with patch("win32api.GetLocalTime", MagicMock(return_value=get_local_time)), patch(
        "ctypes.windll.kernel32.SetLocalTime", MagicMock()
    ) as set_local_time:
        ret = win_system.set_system_time("12:05 AM")
        assert ret


def test_get_system_date():
    """
    Test setting the system date
    """
    ret = win_system.get_system_date()
    assert ret


def test_set_system_date():
    """
    Test setting the system time
    """
    with patch.object(win_system, "set_system_date_time", MagicMock()):
        ret = win_system.set_system_date("03-28-13")
        assert ret



"""
Test the iosconfig Execution module.
"""

import textwrap
from collections import OrderedDict

import pytest

import salt.modules.iosconfig as iosconfig


@pytest.fixture
def running_config():
    return textwrap.dedent(
        """\
    interface GigabitEthernet1
     ip address dhcp
     negotiation auto
     no mop enabled
    !
    interface GigabitEthernet2
     ip address 172.20.0.1 255.255.255.0
     shutdown
     negotiation auto
    !
    interface GigabitEthernet3
     no ip address
     shutdown
     negotiation auto
    !"""
    )


@pytest.fixture
def candidate_config():
    return textwrap.dedent(
        """\
    interface GigabitEthernet1
     ip address dhcp
     negotiation auto
     no mop enabled
    !
    interface GigabitEthernet2
     no ip address
     shutdown
     negotiation auto
    !
    interface GigabitEthernet3
     no ip address
     negotiation auto
    !
    router bgp 65000
     bgp log-neighbor-changes
     neighbor 1.1.1.1 remote-as 12345
    !
    !"""
    )


@pytest.fixture
def merge_config():
    return textwrap.dedent(
        """\
    router bgp 65000
     bgp log-neighbor-changes
     neighbor 1.1.1.1 remote-as 12345
    !
    !
    virtual-service csr_mgmt
    !
    ip forward-protocol nd
    !"""
    )


@pytest.fixture
def configure_loader_modules():
    return {iosconfig: {}}


def test_tree(running_config):
    running_config_tree = OrderedDict(
        [
            (
                "interface GigabitEthernet1",
                OrderedDict(
                    [
                        ("ip address dhcp", OrderedDict()),
                        ("negotiation auto", OrderedDict()),
                        ("no mop enabled", OrderedDict()),
                    ]
                ),
            ),
            (
                "interface GigabitEthernet2",
                OrderedDict(
                    [
                        ("ip address 172.20.0.1 255.255.255.0", OrderedDict()),
                        ("shutdown", OrderedDict()),
                        ("negotiation auto", OrderedDict()),
                    ]
                ),
            ),
            (
                "interface GigabitEthernet3",
                OrderedDict(
                    [
                        ("no ip address", OrderedDict()),
                        ("shutdown", OrderedDict()),
                        ("negotiation auto", OrderedDict()),
                    ]
                ),
            ),
        ]
    )
    tree = iosconfig.tree(config=running_config)
    assert tree == running_config_tree


def test_clean(running_config):
    clean_running_config = textwrap.dedent(
        """\
        interface GigabitEthernet1
         ip address dhcp
         negotiation auto
         no mop enabled
        interface GigabitEthernet2
         ip address 172.20.0.1 255.255.255.0
         shutdown
         negotiation auto
        interface GigabitEthernet3
         no ip address
         shutdown
         negotiation auto
    """
    )
    clean = iosconfig.clean(config=running_config)
    assert clean == clean_running_config


def test_merge_tree(running_config, merge_config):
    expected_merge_tree = OrderedDict(
        [
            (
                "interface GigabitEthernet1",
                OrderedDict(
                    [
                        ("ip address dhcp", OrderedDict()),
                        ("negotiation auto", OrderedDict()),
                        ("no mop enabled", OrderedDict()),
                    ]
                ),
            ),
            (
                "interface GigabitEthernet2",
                OrderedDict(
                    [
                        ("ip address 172.20.0.1 255.255.255.0", OrderedDict()),
                        ("shutdown", OrderedDict()),
                        ("negotiation auto", OrderedDict()),
                    ]
                ),
            ),
            (
                "interface GigabitEthernet3",
                OrderedDict(
                    [
                        ("no ip address", OrderedDict()),
                        ("shutdown", OrderedDict()),
                        ("negotiation auto", OrderedDict()),
                    ]
                ),
            ),
            (
                "router bgp 65000",
                OrderedDict(
                    [
                        ("bgp log-neighbor-changes", OrderedDict()),
                        ("neighbor 1.1.1.1 remote-as 12345", OrderedDict()),
                    ]
                ),
            ),
            ("virtual-service csr_mgmt", OrderedDict()),
            ("ip forward-protocol nd", OrderedDict()),
        ]
    )
    merge_tree = iosconfig.merge_tree(
        initial_config=running_config, merge_config=merge_config
    )
    assert merge_tree == expected_merge_tree


def test_merge_text(running_config, merge_config):
    expected_merge_text = textwrap.dedent(
        """\
        interface GigabitEthernet1
         ip address dhcp
         negotiation auto
         no mop enabled
        interface GigabitEthernet2
         ip address 172.20.0.1 255.255.255.0
         shutdown
         negotiation auto
        interface GigabitEthernet3
         no ip address
         shutdown
         negotiation auto
        router bgp 65000
         bgp log-neighbor-changes
         neighbor 1.1.1.1 remote-as 12345
        virtual-service csr_mgmt
        ip forward-protocol nd
    """
    )
    merge_text = iosconfig.merge_text(
        initial_config=running_config, merge_config=merge_config
    )
    assert merge_text == expected_merge_text


def test_merge_diff(running_config, merge_config):
    expected_diff = textwrap.dedent(
        """\
        @@ -10,3 +10,8 @@
          no ip address
          shutdown
          negotiation auto
        +router bgp 65000
        + bgp log-neighbor-changes
        + neighbor 1.1.1.1 remote-as 12345
        +virtual-service csr_mgmt
        +ip forward-protocol nd
    """
    )
    diff = iosconfig.merge_diff(
        initial_config=running_config, merge_config=merge_config
    )
    assert diff.splitlines()[2:] == expected_diff.splitlines()


def test_diff_text(running_config, candidate_config):
    expected_diff = textwrap.dedent(
        """\
        @@ -3,10 +3,12 @@
          negotiation auto
          no mop enabled
         interface GigabitEthernet2
        - ip address 172.20.0.1 255.255.255.0
        + no ip address
          shutdown
          negotiation auto
         interface GigabitEthernet3
          no ip address
        - shutdown
          negotiation auto
        +router bgp 65000
        + bgp log-neighbor-changes
        + neighbor 1.1.1.1 remote-as 12345
        """
    )
    diff = iosconfig.diff_text(
        candidate_config=candidate_config, running_config=running_config
    )
    assert diff.splitlines()[2:] == expected_diff.splitlines()



"""
Microsoft Update files management via wusa.exe

:maintainer:    Thomas Lemarchand
:platform:      Windows
:depends:       PowerShell

.. versionadded:: 2018.3.4
"""

import logging
import os

import salt.utils.platform
from salt.exceptions import CommandExecutionError

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "wusa"


def __virtual__():
    """
    Load only on Windows
    """
    if not salt.utils.platform.is_windows():
        return False, "Only available on Windows systems"

    powershell_info = __salt__["cmd.shell_info"](shell="powershell", list_modules=False)
    if not powershell_info["installed"]:
        return False, "PowerShell not available"

    return __virtualname__


def _pshell_json(cmd, cwd=None):
    """
    Execute the desired powershell command and ensure that it returns data
    in JSON format and load that into python
    """
    if "convertto-json" not in cmd.lower():
        cmd = f"{cmd} | ConvertTo-Json"
    log.debug("PowerShell: %s", cmd)
    ret = __salt__["cmd.run_all"](cmd, shell="powershell", cwd=cwd)

    if "pid" in ret:
        del ret["pid"]

    if ret.get("stderr", ""):
        error = ret["stderr"].splitlines()[0]
        raise CommandExecutionError(error, info=ret)

    if "retcode" not in ret or ret["retcode"] != 0:
        # run_all logs an error to log.error, fail hard back to the user
        raise CommandExecutionError(f"Issue executing PowerShell {cmd}", info=ret)

    # Sometimes Powershell returns an empty string, which isn't valid JSON
    if ret["stdout"] == "":
        ret["stdout"] = "{}"

    try:
        ret = salt.utils.json.loads(ret["stdout"], strict=False)
    except ValueError:
        raise CommandExecutionError("No JSON results from PowerShell", info=ret)

    return ret


def is_installed(name):
    """
    Check if a specific KB is installed.

    Args:

        name (str):
            The name of the KB to check.

    Returns:
        bool: ``True`` if installed, otherwise ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' wusa.is_installed KB123456
    """
    return (
        __salt__["cmd.retcode"](
            cmd=f"Get-HotFix -Id {name}",
            shell="powershell",
            ignore_retcode=True,
        )
        == 0
    )


def install(path, restart=False):
    """
    Install a KB from a .msu file.

    Args:

        path (str):
            The full path to the msu file to install

        restart (:obj:`bool`, optional):
            ``True`` to force a restart if required by the installation. Adds
            the ``/forcerestart`` switch to the ``wusa.exe`` command. ``False``
            will add the ``/norestart`` switch instead.

            Default is ``False``.

    Returns:
        bool: ``True`` if successful, otherwise ``False``.

    Raise:
        CommandExecutionError: If the package is already installed or an error
            is encountered

    CLI Example:

    .. code-block:: bash

        salt '*' wusa.install C:/temp/KB123456.msu
    """
    # Build the command
    cmd = ["wusa.exe", path, "/quiet"]
    if restart:
        cmd.append("/forcerestart")
    else:
        cmd.append("/norestart")

    # Run the command
    ret_code = __salt__["cmd.retcode"](cmd, ignore_retcode=True)

    # Check the ret_code
    file_name = os.path.basename(path)
    errors = {
        2359302: f"{file_name} is already installed",
        3010: (
            "{} correctly installed but server reboot is needed to complete"
            " installation".format(file_name)
        ),
        87: "Unknown error",
    }
    if ret_code in errors:
        raise CommandExecutionError(errors[ret_code], ret_code)
    elif ret_code:
        raise CommandExecutionError(f"Unknown error: {ret_code}")

    return True


def uninstall(path, restart=False):
    """
    Uninstall a specific KB.

    Args:

        path (str):
            The full path to the msu file to uninstall. This can also be just
            the name of the KB to uninstall

        restart (:obj:`bool`, optional):
            ``True`` to force a restart if required by the installation. Adds
            the ``/forcerestart`` switch to the ``wusa.exe`` command. ``False``
            will add the ``/norestart`` switch instead.

            Default is ``False``.

    Returns:
        bool: ``True`` if successful, otherwise ``False``.

    Raises:
        CommandExecutionError: If an error is encountered

    CLI Example:

    .. code-block:: bash

        salt '*' wusa.uninstall KB123456

        # or

        salt '*' wusa.uninstall C:/temp/KB123456.msu
    """
    # Build the command
    cmd = ["wusa.exe", "/uninstall", "/quiet"]
    kb = os.path.splitext(os.path.basename(path))[0]
    if os.path.exists(path):
        cmd.append(path)
    else:
        cmd.append("/kb:{}".format(kb[2:] if kb.lower().startswith("kb") else kb))
    if restart:
        cmd.append("/forcerestart")
    else:
        cmd.append("/norestart")

    # Run the command
    ret_code = __salt__["cmd.retcode"](cmd, ignore_retcode=True)

    # Check the ret_code
    # If you pass /quiet and specify /kb, you'll always get retcode 87 if there
    # is an error. Use the actual file to get a more descriptive error
    errors = {
        -2145116156: f"{kb} does not support uninstall",
        2359303: f"{kb} not installed",
        87: "Unknown error. Try specifying an .msu file",
    }
    if ret_code in errors:
        raise CommandExecutionError(errors[ret_code], ret_code)
    elif ret_code:
        raise CommandExecutionError(f"Unknown error: {ret_code}")

    return True


def list():
    """
    Get a list of updates installed on the machine.

    Returns:
        list: A list of installed updates.

    CLI Example:

    .. code-block:: bash

        salt '*' wusa.list
    """
    kbs = []
    ret = _pshell_json("Get-HotFix | Select HotFixID")
    for item in ret:
        kbs.append(item["HotFixID"])
    return kbs



"""
Salt performance tests
"""

import logging
import shutil

import pytest
from saltfactories.utils import random_string

from tests.support.sminion import create_sminion

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_binaries_missing("docker"),
]


log = logging.getLogger(__name__)


@pytest.fixture(scope="session")
def docker_network_name():
    return random_string("salt-perf-", uppercase=False)


@pytest.fixture(scope="session")
def host_docker_network_ip_address(docker_network_name):
    sminion = create_sminion()
    network_subnet = "10.0.21.0/24"
    network_gateway = "10.0.21.1"
    try:
        ret = sminion.states.docker_network.present(
            docker_network_name,
            driver="bridge",
            ipam_pools=[{"subnet": network_subnet, "gateway": network_gateway}],
        )
        assert isinstance(ret, dict), ret
        try:
            assert ret["result"]
        except AssertionError:
            pytest.skip(f"Failed to create docker network: {ret}")
        yield network_gateway
    finally:
        sminion.states.docker_network.absent(docker_network_name)


@pytest.fixture(scope="session")
def salt_factories_config(salt_factories_config, host_docker_network_ip_address):
    """
    Return a dictionary with the keyword arguments for FactoriesManager
    """
    config = salt_factories_config.copy()
    config["log_server_host"] = host_docker_network_ip_address
    return config


@pytest.fixture(scope="package")
def integration_files_dir(tmp_path_factory):
    """
    Fixture which returns the salt integration files directory path.
    Creates the directory if it does not yet exist.
    """
    dirname = tmp_path_factory.mktemp("integration-files")
    try:
        yield dirname
    finally:
        shutil.rmtree(str(dirname), ignore_errors=True)


@pytest.fixture(scope="package")
def state_tree(integration_files_dir):
    """
    Fixture which returns the salt state tree root directory path.
    Creates the directory if it does not yet exist.
    """
    dirname = integration_files_dir / "state-tree"
    dirname.mkdir(exist_ok=True)
    (dirname / "testfile").write_text("This is a test file")
    return dirname


@pytest.fixture(scope="package")
def sls_contents():
    return """
    add_file:
      file.managed:
        - name: {path}
        - source: salt://testfile
        - makedirs: true
        - require:
          - cmd: echo
    delete_file:
      file.absent:
        - name: {path}
        - require:
          - file: add_file
    echo:
      cmd.run:
        - name: \"echo 'This is a test!'\"
    """


@pytest.fixture(scope="package")
def file_add_delete_sls(testfile_path, state_tree):
    sls_name = "file_add"
    sls_contents = """
    add_file:
      file.managed:
        - name: {path}
        - source: salt://testfile
        - makedirs: true
        - require:
          - cmd: echo
    delete_file:
      file.absent:
        - name: {path}
        - require:
          - file: add_file
    echo:
      cmd.run:
        - name: \"echo 'This is a test!'\"
    """.format(
        path=testfile_path
    )
    with pytest.helpers.temp_file(f"{sls_name}.sls", sls_contents, state_tree):
        yield sls_name



#!/usr/bin/env python

"""
The minionswarm script will start a group of salt minions with different ids
on a single system to test scale capabilities
"""
# pylint: disable=resource-leakage

import hashlib
import optparse  # pylint: disable=deprecated-module
import os
import random
import shutil
import signal
import subprocess
import sys
import tempfile
import time
import uuid

import salt
import salt.utils.files
import salt.utils.yaml
import tests.support.runtests

OSES = [
    "Arch",
    "Ubuntu",
    "Debian",
    "CentOS",
    "Fedora",
    "Gentoo",
    "AIX",
    "Solaris",
]
VERS = [
    "2014.1.6",
    "2014.7.4",
    "2015.5.5",
    "2015.8.0",
]


def parse():
    """
    Parse the cli options
    """
    parser = optparse.OptionParser()
    parser.add_option(
        "-m",
        "--minions",
        dest="minions",
        default=5,
        type="int",
        help="The number of minions to make",
    )
    parser.add_option(
        "-M",
        action="store_true",
        dest="master_too",
        default=False,
        help="Run a local master and tell the minions to connect to it",
    )
    parser.add_option(
        "--master",
        dest="master",
        default="salt",
        help="The location of the salt master that this swarm will serve",
    )
    parser.add_option(
        "--name",
        "-n",
        dest="name",
        default="ms",
        help=(
            "Give the minions an alternative id prefix, this is used "
            "when minions from many systems are being aggregated onto "
            "a single master"
        ),
    )
    parser.add_option(
        "--rand-os",
        dest="rand_os",
        default=False,
        action="store_true",
        help="Each Minion claims a different os grain",
    )
    parser.add_option(
        "--rand-ver",
        dest="rand_ver",
        default=False,
        action="store_true",
        help="Each Minion claims a different version grain",
    )
    parser.add_option(
        "--rand-machine-id",
        dest="rand_machine_id",
        default=False,
        action="store_true",
        help="Each Minion claims a different machine id grain",
    )
    parser.add_option(
        "--rand-uuid",
        dest="rand_uuid",
        default=False,
        action="store_true",
        help="Each Minion claims a different UUID grain",
    )
    parser.add_option(
        "-k",
        "--keep-modules",
        dest="keep",
        default="",
        help="A comma delimited list of modules to enable",
    )
    parser.add_option(
        "-f",
        "--foreground",
        dest="foreground",
        default=False,
        action="store_true",
        help="Run the minions with debug output of the swarm going to the terminal",
    )
    parser.add_option(
        "--temp-dir",
        dest="temp_dir",
        default=None,
        help="Place temporary files/directories here",
    )
    parser.add_option(
        "--no-clean",
        action="store_true",
        default=False,
        help="Don't cleanup temporary files/directories",
    )
    parser.add_option(
        "--root-dir",
        dest="root_dir",
        default=None,
        help="Override the minion root_dir config",
    )
    parser.add_option(
        "--transport",
        dest="transport",
        default="zeromq",
        help="Declare which transport to use, default is zeromq",
    )
    parser.add_option(
        "--start-delay",
        dest="start_delay",
        default=0.0,
        type="float",
        help="Seconds to wait between minion starts",
    )
    parser.add_option(
        "-c",
        "--config-dir",
        default="",
        help="Pass in a configuration directory containing base configuration.",
    )
    parser.add_option("-u", "--user", default=tests.support.runtests.this_user())

    options, _args = parser.parse_args()

    opts = {}

    for key, val in options.__dict__.items():
        opts[key] = val

    return opts


class Swarm:
    """
    Create a swarm of minions
    """

    def __init__(self, opts):
        self.opts = opts

        # If given a temp_dir, use it for temporary files
        if opts["temp_dir"]:
            self.swarm_root = opts["temp_dir"]
        else:
            # If given a root_dir, keep the tmp files there as well
            if opts["root_dir"]:
                tmpdir = os.path.join(opts["root_dir"], "tmp")
            else:
                tmpdir = opts["root_dir"]
            self.swarm_root = tempfile.mkdtemp(
                prefix="mswarm-root", suffix=".d", dir=tmpdir
            )

        if self.opts["transport"] == "zeromq":
            self.pki = self._pki_dir()
        self.zfill = len(str(self.opts["minions"]))

        self.confs = set()

        random.seed(0)

    def _pki_dir(self):
        """
        Create the shared pki directory
        """
        path = os.path.join(self.swarm_root, "pki")
        if not os.path.exists(path):
            os.makedirs(path)

            print(f"Creating shared pki keys for the swarm on: {path}")
            subprocess.call(
                "salt-key -c {0} --gen-keys minion --gen-keys-dir {0} "
                "--log-file {1} --user {2}".format(
                    path,
                    os.path.join(path, "keys.log"),
                    self.opts["user"],
                ),
                shell=True,
            )
            print("Keys generated")
        return path

    def start(self):
        """
        Start the magic!!
        """
        if self.opts["master_too"]:
            master_swarm = MasterSwarm(self.opts)
            master_swarm.start()
        minions = MinionSwarm(self.opts)
        minions.start_minions()
        print("Starting minions...")
        # self.start_minions()
        print("All {} minions have started.".format(self.opts["minions"]))
        print("Waiting for CTRL-C to properly shutdown minions...")
        while True:
            try:
                time.sleep(5)
            except KeyboardInterrupt:
                print("\nShutting down minions")
                self.clean_configs()
                break

    def shutdown(self):
        """
        Tear it all down
        """
        print("Killing any remaining running minions")
        subprocess.call('pkill -KILL -f "python.*salt-minion"', shell=True)
        if self.opts["master_too"]:
            print("Killing any remaining masters")
            subprocess.call('pkill -KILL -f "python.*salt-master"', shell=True)
        if not self.opts["no_clean"]:
            print("Remove ALL related temp files/directories")
            shutil.rmtree(self.swarm_root)
        print("Done")

    def clean_configs(self):
        """
        Clean up the config files
        """
        for path in self.confs:
            pidfile = f"{path}.pid"
            try:
                try:
                    with salt.utils.files.fopen(pidfile) as fp_:
                        pid = int(fp_.read().strip())
                    os.kill(pid, signal.SIGTERM)
                except ValueError:
                    pass
                if os.path.exists(pidfile):
                    os.remove(pidfile)
                if not self.opts["no_clean"]:
                    shutil.rmtree(path)
            except OSError:
                pass


class MinionSwarm(Swarm):
    """
    Create minions
    """

    def start_minions(self):
        """
        Iterate over the config files and start up the minions
        """
        self.prep_configs()
        for path in self.confs:
            cmd = "salt-minion -c {} --pid-file {}".format(path, f"{path}.pid")
            if self.opts["foreground"]:
                cmd += " -l debug &"
            else:
                cmd += " -d &"
            subprocess.call(cmd, shell=True)
            time.sleep(self.opts["start_delay"])

    def mkconf(self, idx):
        """
        Create a config file for a single minion
        """
        data = {}
        if self.opts["config_dir"]:
            spath = os.path.join(self.opts["config_dir"], "minion")
            with salt.utils.files.fopen(spath) as conf:
                data = salt.utils.yaml.safe_load(conf) or {}
        minion_id = "{}-{}".format(self.opts["name"], str(idx).zfill(self.zfill))

        dpath = os.path.join(self.swarm_root, minion_id)
        if not os.path.exists(dpath):
            os.makedirs(dpath)

        data.update(
            {
                "id": minion_id,
                "user": self.opts["user"],
                "cachedir": os.path.join(dpath, "cache"),
                "master": self.opts["master"],
                "log_file": os.path.join(dpath, "minion.log"),
                "grains": {},
            }
        )

        if self.opts["transport"] == "zeromq":
            minion_pkidir = os.path.join(dpath, "pki")
            if not os.path.exists(minion_pkidir):
                os.makedirs(minion_pkidir)
                minion_pem = os.path.join(self.pki, "minion.pem")
                minion_pub = os.path.join(self.pki, "minion.pub")
                shutil.copy(minion_pem, minion_pkidir)
                shutil.copy(minion_pub, minion_pkidir)
            data["pki_dir"] = minion_pkidir
        elif self.opts["transport"] == "tcp":
            data["transport"] = "tcp"

        if self.opts["root_dir"]:
            data["root_dir"] = self.opts["root_dir"]

        path = os.path.join(dpath, "minion")

        if self.opts["keep"]:
            keep = self.opts["keep"].split(",")
            modpath = os.path.join(os.path.dirname(salt.__file__), "modules")
            fn_prefixes = (fn_.partition(".")[0] for fn_ in os.listdir(modpath))
            ignore = [fn_prefix for fn_prefix in fn_prefixes if fn_prefix not in keep]
            data["disable_modules"] = ignore

        if self.opts["rand_os"]:
            data["grains"]["os"] = random.choice(OSES)
        if self.opts["rand_ver"]:
            data["grains"]["saltversion"] = random.choice(VERS)
        if self.opts["rand_machine_id"]:
            data["grains"]["machine_id"] = hashlib.md5(minion_id).hexdigest()
        if self.opts["rand_uuid"]:
            data["grains"]["uuid"] = str(uuid.uuid4())

        with salt.utils.files.fopen(path, "w+") as fp_:
            salt.utils.yaml.safe_dump(data, fp_)
        self.confs.add(dpath)

    def prep_configs(self):
        """
        Prepare the confs set
        """
        for idx in range(self.opts["minions"]):
            self.mkconf(idx)


class MasterSwarm(Swarm):
    """
    Create one or more masters
    """

    def __init__(self, opts):
        super().__init__(opts)
        self.conf = os.path.join(self.swarm_root, "master")

    def start(self):
        """
        Prep the master start and fire it off
        """
        # sys.stdout for no newline
        sys.stdout.write("Generating master config...")
        self.mkconf()
        print("done")

        sys.stdout.write("Starting master...")
        self.start_master()
        print("done")

    def start_master(self):
        """
        Do the master start
        """
        cmd = "salt-master -c {} --pid-file {}".format(self.conf, f"{self.conf}.pid")
        if self.opts["foreground"]:
            cmd += " -l debug &"
        else:
            cmd += " -d &"
        subprocess.call(cmd, shell=True)

    def mkconf(self):  # pylint: disable=W0221
        """
        Make a master config and write it'
        """
        data = {}
        if self.opts["config_dir"]:
            spath = os.path.join(self.opts["config_dir"], "master")
            with salt.utils.files.fopen(spath) as conf:
                data = salt.utils.yaml.safe_load(conf)
        data.update(
            {
                "log_file": os.path.join(self.conf, "master.log"),
                "open_mode": True,  # TODO Pre-seed keys
            }
        )

        os.makedirs(self.conf)
        path = os.path.join(self.conf, "master")

        with salt.utils.files.fopen(path, "w+") as fp_:
            salt.utils.yaml.safe_dump(data, fp_)

    def shutdown(self):
        print("Killing master")
        subprocess.call('pkill -KILL -f "python.*salt-master"', shell=True)
        print("Master killed")


# pylint: disable=C0103
if __name__ == "__main__":
    swarm = Swarm(parse())
    try:
        swarm.start()
    finally:
        swarm.shutdown()



"""
Support for scheduling celery tasks. The worker is independent of salt and thus can run in a different
virtualenv or on a different python version, as long as broker, backend and serializer configurations match.
Also note that celery and packages required by the celery broker, e.g. redis must be installed to load
the salt celery execution module.

.. note::
    A new app (and thus new connections) is created for each task execution
"""

import logging

from salt.exceptions import SaltInvocationError

log = logging.getLogger(__name__)


try:
    from celery import Celery
    from celery.exceptions import TimeoutError  # pylint: disable=no-name-in-module

    HAS_CELERY = True
except ImportError:
    HAS_CELERY = False


def __virtual__():
    """
    Only load if celery libraries exist.
    """
    if not HAS_CELERY:
        return False, "The celery module could not be loaded: celery library not found"
    return True


def run_task(
    task_name,
    args=None,
    kwargs=None,
    broker=None,
    backend=None,
    wait_for_result=False,
    timeout=None,
    propagate=True,
    interval=0.5,
    no_ack=True,
    raise_timeout=True,
    config=None,
):
    """
    Execute celery tasks. For celery specific parameters see celery documentation.

    CLI Example:

    .. code-block:: bash

        salt '*' celery.run_task tasks.sleep args=[4] broker=redis://localhost \\
        backend=redis://localhost wait_for_result=true

    task_name
        The task name, e.g. tasks.sleep

    args
        Task arguments as a list

    kwargs
        Task keyword arguments

    broker
        Broker for celeryapp, see celery documentation

    backend
        Result backend for celeryapp, see celery documentation

    wait_for_result
        Wait until task result is read from result backend and return result, Default: False

    timeout
        Timeout waiting for result from celery, see celery AsyncResult.get documentation

    propagate
        Propagate exceptions from celery task, see celery AsyncResult.get documentation, Default: True

    interval
        Interval to check for task result, see celery AsyncResult.get documentation, Default: 0.5

    no_ack
        see celery AsyncResult.get documentation. Default: True

    raise_timeout
        Raise timeout exception if waiting for task result times out. Default: False

    config
        Config dict for celery app, See celery documentation

    """
    if not broker:
        raise SaltInvocationError("broker parameter is required")

    with Celery(broker=broker, backend=backend, set_as_current=False) as app:
        if config:
            app.conf.update(config)

        with app.connection():
            args = args or []
            kwargs = kwargs or {}
            async_result = app.send_task(task_name, args=args, kwargs=kwargs)

            if wait_for_result:
                try:
                    return async_result.get(
                        timeout=timeout,
                        propagate=propagate,
                        interval=interval,
                        no_ack=no_ack,
                    )
                except TimeoutError as ex:
                    log.error(
                        "Waiting for the result of a celery task execution timed out."
                    )
                    if raise_timeout:
                        raise
                    return False



"""
Tests for the idem state
"""

import tempfile
from contextlib import contextmanager

import pytest

import salt.state
import salt.utils.idem as idem
import salt.utils.path
import tests.support.sminion

pytestmark = [
    pytest.mark.skipif(not idem.HAS_POP[0], reason=idem.HAS_POP[1]),
]


@contextmanager
def test_state(salt_call_cli):
    with tempfile.NamedTemporaryFile(suffix=".sls", delete=True, mode="w+") as fh:
        sls_succeed_without_changes = """
        state_name:
          test.succeed_without_changes:
            - name: idem_test
            - foo: bar
        """
        fh.write(sls_succeed_without_changes)
        fh.flush()
        ret = salt_call_cli.run(
            "--local", "state.single", "idem.state", sls=fh.name, name="idem_test"
        )

    state_id = "idem_|-idem_test_|-idem_test_|-state"
    parent = ret.data[state_id]
    assert parent["result"] is True, parent["comment"]
    sub_state_ret = parent["sub_state_run"][0]
    assert sub_state_ret["result"] is True
    assert sub_state_ret["name"] == "idem_test"
    assert "Success!" in sub_state_ret["comment"]

    # Format the idem state through the state outputter
    minion_opts = tests.support.sminion.build_minion_opts()
    state_obj = salt.state.State(minion_opts)

    chunk_ret = state_obj.call_chunk(
        {"state": "state", "name": "name", "fun": "fun", "__id__": "__id__"},
        ret.data,
        {},
    )
    # Verify that the sub_state_run looks like a normal salt state
    assert "start_time" in chunk_ret[state_id]
    float(chunk_ret[state_id]["duration"])


def test_bad_state(salt_call_cli):
    bad_sls = "non-existant-file.sls"

    ret = salt_call_cli.run(
        "--local", "state.single", "idem.state", sls=bad_sls, name="idem_bad_test"
    )
    parent = ret.data["idem_|-idem_bad_test_|-idem_bad_test_|-state"]

    assert parent["result"] is False
    assert f"SLS ref {bad_sls} did not resolve to a file" == parent["comment"]
    assert not parent["sub_state_run"]



import ctypes
import logging
import multiprocessing
import pathlib
import shutil
import time

import pytest
from pytestshellutils.utils.processes import terminate_process

import salt.channel.client
import salt.channel.server
import salt.config
import salt.crypt
import salt.exceptions
import salt.ext.tornado.gen
import salt.master
import salt.utils.platform
import salt.utils.process
import salt.utils.stringutils

log = logging.getLogger(__name__)


pytestmark = [
    pytest.mark.skip_on_spawning_platform(
        reason="These tests are currently broken on spawning platforms. Need to be rewritten.",
    ),
    pytest.mark.slow_test,
    pytest.mark.skipif(
        "grains['osfinger'] == 'Rocky Linux-8' and grains['osarch'] == 'aarch64'",
        reason="Temporarily skip on Rocky Linux 8 Arm64",
    ),
]


class ReqServerChannelProcess(salt.utils.process.SignalHandlingProcess):

    def __init__(self, config, req_channel_crypt):
        super().__init__()
        self._closing = False
        self.config = config
        self.req_channel_crypt = req_channel_crypt
        self.process_manager = salt.utils.process.ProcessManager(
            name="ReqServer-ProcessManager"
        )
        self.req_server_channel = salt.channel.server.ReqServerChannel.factory(
            self.config
        )
        self.req_server_channel.pre_fork(self.process_manager)
        self.io_loop = None
        self.running = multiprocessing.Event()

    def run(self):
        salt.master.SMaster.secrets["aes"] = {
            "secret": multiprocessing.Array(
                ctypes.c_char,
                salt.utils.stringutils.to_bytes(
                    salt.crypt.Crypticle.generate_key_string()
                ),
            ),
            "serial": multiprocessing.Value(
                ctypes.c_longlong, lock=False  # We'll use the lock from 'secret'
            ),
        }

        self.io_loop = salt.ext.tornado.ioloop.IOLoop()
        self.io_loop.make_current()
        self.req_server_channel.post_fork(self._handle_payload, io_loop=self.io_loop)
        self.io_loop.add_callback(self.running.set)
        try:
            self.io_loop.start()
        except KeyboardInterrupt:
            pass

    def _handle_signals(self, signum, sigframe):
        self.close()
        super()._handle_signals(signum, sigframe)

    def __enter__(self):
        self.start()
        self.running.wait()
        return self

    def __exit__(self, *args):
        self.close()
        self.terminate()

    def close(self):
        if self._closing:
            return
        self._closing = True
        if self.req_server_channel is not None:
            self.req_server_channel.close()
            self.req_server_channel = None
        if self.process_manager is not None:
            self.process_manager.terminate()
            # Really terminate any process still left behind
            for pid in self.process_manager._process_map:
                terminate_process(pid=pid, kill_children=True, slow_stop=False)
            self.process_manager = None

    @salt.ext.tornado.gen.coroutine
    def _handle_payload(self, payload):
        if self.req_channel_crypt == "clear":
            raise salt.ext.tornado.gen.Return((payload, {"fun": "send_clear"}))
        raise salt.ext.tornado.gen.Return((payload, {"fun": "send"}))


@pytest.fixture
def req_server_channel(salt_master, req_channel_crypt):
    req_server_channel_process = ReqServerChannelProcess(
        salt_master.config.copy(), req_channel_crypt
    )
    try:
        with req_server_channel_process:
            yield
    finally:
        terminate_process(
            pid=req_server_channel_process.pid, kill_children=True, slow_stop=False
        )


@pytest.fixture
def req_server_opts(tmp_path):
    sock_dir = tmp_path / "sock"
    pki_dir = tmp_path / "pki"
    cache_dir = tmp_path / "cache"
    sock_dir.mkdir()
    pki_dir.mkdir()
    cache_dir.mkdir()
    yield {
        "sock_dir": sock_dir,
        "pki_dir": pki_dir,
        "cachedir": cache_dir,
        "key_pass": "meh",
        "keysize": 2048,
        "cluster_id": None,
        "master_sign_pubkey": False,
        "pub_server_niceness": None,
        "con_cache": False,
        "zmq_monitor": False,
        "request_server_ttl": 60,
        "publish_session": 600,
    }


@pytest.fixture
def req_server(req_server_opts):
    server = salt.channel.server.ReqServerChannel.factory(req_server_opts)
    try:
        yield server
    finally:
        server.close()


@pytest.fixture
def minion1_id():
    yield "minion1"


@pytest.fixture
def minion1_key(minion1_id, tmp_path, req_server_opts):
    minionpki = tmp_path / minion1_id
    minionpki.mkdir()
    key1 = pathlib.Path(salt.crypt.gen_keys(minionpki, minion1_id, 2048))

    pki = pathlib.Path(req_server_opts["pki_dir"])
    (pki / "minions").mkdir(exist_ok=True)
    shutil.copy2(key1.with_suffix(".pub"), pki / "minions" / minion1_id)
    yield salt.crypt.PrivateKey(key1)


@pytest.fixture
def minion2_id():
    yield "minion2"


@pytest.fixture
def minion2_key(minion2_id, tmp_path, req_server_opts):
    minionpki = tmp_path / minion2_id
    minionpki.mkdir()
    key2 = pathlib.Path(salt.crypt.gen_keys(minionpki, minion2_id, 2048))

    pki = pathlib.Path(req_server_opts["pki_dir"])
    (pki / "minions").mkdir(exist_ok=True)
    shutil.copy2(key2.with_suffix(".pub"), pki / "minions" / minion2_id)
    yield salt.crypt.PrivateKey(key2)


def req_channel_crypt_ids(value):
    return f"ReqChannel(crypt='{value}')"


@pytest.fixture(params=["clear", "aes"], ids=req_channel_crypt_ids)
def req_channel_crypt(request):
    return request.param


@pytest.fixture
def push_channel(req_server_channel, salt_minion, req_channel_crypt):
    with salt.channel.client.ReqChannel.factory(
        salt_minion.config, crypt=req_channel_crypt
    ) as _req_channel:
        try:
            yield _req_channel
        finally:
            # Force termination of singleton
            _req_channel.obj._refcount = 0


def test_basic(push_channel):
    """
    Test a variety of messages, make sure we get the expected responses
    """
    if push_channel.crypt == "aes":
        pytest.skip(reason="test not valid for encrypted channel")
    msgs = [
        {"foo": "bar"},
        {"bar": "baz"},
        {"baz": "qux", "list": [1, 2, 3]},
    ]

    for msg in msgs:
        ret = push_channel.send(dict(msg), timeout=5, tries=1)
        assert ret["load"] == msg


def test_normalization(push_channel):
    """
    Since we use msgpack, we need to test that list types are converted to lists
    """
    if push_channel.crypt == "aes":
        pytest.skip(reason="test not valid for encrypted channel")
    types = {
        "list": list,
    }
    msgs = [
        {"list": tuple([1, 2, 3])},
    ]
    for msg in msgs:
        ret = push_channel.send(msg, timeout=5, tries=1)
        for key, value in ret["load"].items():
            assert types[key] == type(value)


def test_badload(push_channel, req_channel_crypt):
    """
    Test a variety of bad requests, make sure that we get some sort of error
    """
    if push_channel.crypt == "aes":
        pytest.skip(reason="test not valid for encrypted channel")
    msgs = ["", [], tuple()]
    if req_channel_crypt == "clear":
        for msg in msgs:
            ret = push_channel.send(msg, timeout=5, tries=1)
            assert ret == "payload and load must be a dict"
    else:
        for msg in msgs:
            with pytest.raises(salt.exceptions.AuthenticationError):
                push_channel.send(msg, timeout=5, tries=1)


async def test_req_channel_ttl_v2(req_server, io_loop):
    req_server.opts["request_server_ttl"] = 60

    async def handler(payload):
        return payload, {"fun": "send"}

    req_server.post_fork(handler, io_loop)
    payload = {
        "enc": "aes",
        "version": 2,
        "load": req_server.crypticle.dumps(
            {
                "ts": int(time.time() - 61),
            }
        ),
    }
    ret = await req_server.handle_message(payload)
    ret = req_server.crypticle.loads(ret)
    assert ret == payload


async def test_req_channel_ttl_valid(req_server, io_loop, minion1_id, minion1_key):
    req_server.opts["request_server_ttl"] = 60
    req_server.opts["publish_session"] = 600
    tok = minion1_key.encrypt(b"salt")

    async def handler(payload):
        return payload, {"fun": "send"}

    req_server.post_fork(handler, io_loop)
    key = req_server.session_key(minion1_id)

    crypticle = salt.crypt.Crypticle(req_server.opts, key)
    payload = {
        "enc": "aes",
        "id": minion1_id,
        "version": 3,
        "load": crypticle.dumps(
            {
                "ts": int(time.time()),
                "id": minion1_id,
                "tok": tok,
            }
        ),
    }
    ret = await req_server.handle_message(payload)
    ret = crypticle.loads(ret)
    assert ret == payload


async def test_req_channel_ttl_expired(
    req_server, io_loop, caplog, minion1_id, minion1_key
):
    req_server.opts["request_server_ttl"] = 60
    req_server.opts["publish_session"] = 600
    tok = minion1_key.encrypt(b"salt")

    async def handler(payload):
        return payload, {"fun": "send"}

    req_server.post_fork(handler, io_loop)
    key = req_server.session_key(minion1_id)
    crypticle = salt.crypt.Crypticle(req_server.opts, key)
    payload = {
        "enc": "aes",
        "id": minion1_id,
        "version": 3,
        "load": crypticle.dumps(
            {
                "ts": int(time.time() - 61),
                "id": minion1_id,
                "tok": tok,
            }
        ),
    }
    with caplog.at_level(logging.WARNING):
        ret = await req_server.handle_message(payload)
        assert f"Received request from {minion1_id} with expired ttl" in caplog.text
        assert ret == "bad load"


async def test_req_channel_id_invalid_chars(
    req_server, minion1_id, minion1_key, io_loop, caplog
):
    req_server.opts["request_server_ttl"] = 60
    req_server.opts["publish_session"] = 600
    tok = minion1_key.encrypt(b"salt")

    async def handler(payload):
        return payload, {"fun": "send"}

    req_server.post_fork(handler, io_loop)
    key = req_server.session_key(minion1_id)
    crypticle = salt.crypt.Crypticle(req_server.opts, key)
    payload = {
        "enc": "aes",
        "id": f"{minion1_id}\0",
        "version": 3,
        "load": crypticle.dumps(
            {
                "ts": int(time.time()),
                "id": f"{minion1_id}\0",
                "tok": tok,
            }
        ),
    }
    with caplog.at_level(logging.WARNING):
        ret = await req_server.handle_message(payload)
        assert (
            "Bad load from minion: SaltDeserializationError: Encountered invalid id"
            in caplog.text
        )
        assert ret == "bad load"


async def test_req_channel_id_mismatch(
    req_server, io_loop, caplog, minion1_id, minion1_key
):

    id2 = "minion2"

    async def handler(payload):
        return payload, {"fun": "send"}

    req_server.post_fork(handler, io_loop)
    key = req_server.session_key(minion1_id)
    crypticle = salt.crypt.Crypticle(req_server.opts, key)
    payload = {
        "enc": "aes",
        "id": minion1_id,
        "version": 3,
        "load": crypticle.dumps(
            {
                "ts": int(time.time()),
                "id": id2,
            }
        ),
    }
    with caplog.at_level(logging.WARNING):
        ret = await req_server.handle_message(payload)
        assert (
            f"Request id mismatch. Found '{id2}' but expected '{minion1_id}'"
            in caplog.text
        )
        assert ret == "bad load"


async def test_req_channel_v2_invalid_token(
    req_server,
    io_loop,
    caplog,
    tmp_path,
    minion1_id,
    minion1_key,
    minion2_key,
    minion2_id,
):

    tok2 = minion2_key.encrypt(b"salt")

    async def handler(payload):
        return payload, {"fun": "send"}

    req_server.post_fork(handler, io_loop)

    # Minion 1 is trying to impersonate minion2's token.
    payload = {
        "enc": "aes",
        "version": 2,
        "load": req_server.crypticle.dumps(
            {
                "ts": int(time.time()),
                "id": minion1_id,
                "tok": tok2,
            }
        ),
    }
    with caplog.at_level(logging.WARNING):
        ret = await req_server.handle_message(payload)
        assert "Unable to decrypt token:" in caplog.text
        assert ret == "bad load"



import pytest
from saltfactories.utils import random_string

import salt.utils.stringutils
import salt.utils.win_reg as win_reg
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch

try:
    import win32api

except ImportError:
    pass

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
]


@pytest.fixture
def fake_key():
    return "SOFTWARE\\{}".format(random_string("SaltTesting-", lowercase=False))


@pytest.fixture
def unicode_key():
    return "Unicode Key \N{TRADE MARK SIGN}"


@pytest.fixture
def unicode_value():
    return "Unicode Value \N{COPYRIGHT SIGN},\N{TRADE MARK SIGN},\N{REGISTERED SIGN}"


def test_broadcast_change_success():
    """
    Tests the broadcast_change function
    """
    with patch("win32gui.SendMessageTimeout", return_value=("", 0)):
        assert win_reg.broadcast_change()


def test_broadcast_change_fail():
    """
    Tests the broadcast_change function failure
    """
    with patch("win32gui.SendMessageTimeout", return_value=("", 1)):
        assert not win_reg.broadcast_change()


def test_key_exists_existing():
    """
    Tests the key_exists function using a well known registry key
    """
    assert win_reg.key_exists(hive="HKLM", key="SOFTWARE\\Microsoft")


def test_key_exists_non_existing():
    """
    Tests the key_exists function using a non existing registry key
    """
    assert not win_reg.key_exists(hive="HKLM", key="SOFTWARE\\FakeKey")


def test_key_exists_invalid_hive():
    """
    Tests the key_exists function using an invalid hive
    """
    pytest.raises(
        CommandExecutionError,
        win_reg.key_exists,
        hive="BADHIVE",
        key="SOFTWARE\\Microsoft",
    )


def test_key_exists_unknown_key_error():
    """
    Tests the key_exists function with an unknown key error
    """
    mock_error = MagicMock(
        side_effect=win32api.error(123, "RegOpenKeyEx", "Unknown error")
    )
    with patch("salt.utils.win_reg.win32api.RegOpenKeyEx", mock_error):
        pytest.raises(
            win32api.error,
            win_reg.key_exists,
            hive="HKLM",
            key="SOFTWARE\\Microsoft",
        )


def test_value_exists_existing():
    """
    Tests the value_exists function using a well known registry key
    """
    assert win_reg.value_exists(
        hive="HKLM",
        key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion",
        vname="CommonFilesDir",
    )


def test_value_exists_non_existing():
    """
    Tests the value_exists function using a non existing registry key
    """
    assert not win_reg.value_exists(
        hive="HKLM",
        key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion",
        vname="NonExistingValueName",
    )


def test_value_exists_invalid_hive():
    """
    Tests the value_exists function using an invalid hive
    """
    pytest.raises(
        CommandExecutionError,
        win_reg.value_exists,
        hive="BADHIVE",
        key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion",
        vname="CommonFilesDir",
    )


def test_value_exists_key_not_exist():
    """
    Tests the value_exists function when the key does not exist
    """
    mock_error = MagicMock(
        side_effect=win32api.error(2, "RegOpenKeyEx", "Unknown error")
    )
    with patch("salt.utils.win_reg.win32api.RegOpenKeyEx", mock_error):
        assert not win_reg.value_exists(
            hive="HKLM",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion",
            vname="CommonFilesDir",
        )


def test_value_exists_unknown_key_error():
    """
    Tests the value_exists function with an unknown error when opening the
    key
    """
    mock_error = MagicMock(
        side_effect=win32api.error(123, "RegOpenKeyEx", "Unknown error")
    )
    with patch("salt.utils.win_reg.win32api.RegOpenKeyEx", mock_error):
        pytest.raises(
            win32api.error,
            win_reg.value_exists,
            hive="HKLM",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion",
            vname="CommonFilesDir",
        )


def test_value_exists_empty_default_value():
    """
    Tests the value_exists function when querying the default value
    """
    mock_error = MagicMock(
        side_effect=win32api.error(2, "RegQueryValueEx", "Empty Value")
    )
    with patch("salt.utils.win_reg.win32api.RegQueryValueEx", mock_error):
        assert win_reg.value_exists(
            hive="HKLM",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion",
            vname=None,
        )


def test_value_exists_no_vname():
    """
    Tests the value_exists function when the vname does not exist
    """
    mock_error = MagicMock(
        side_effect=win32api.error(123, "RegQueryValueEx", "Empty Value")
    )
    with patch("salt.utils.win_reg.win32api.RegQueryValueEx", mock_error):
        assert not win_reg.value_exists(
            hive="HKLM",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion",
            vname="NonExistingValuePair",
        )


def test_list_keys_existing():
    """
    Test the list_keys function using a well known registry key
    """
    assert "Microsoft" in win_reg.list_keys(hive="HKLM", key="SOFTWARE")


def test_list_keys_non_existing(fake_key):
    """
    Test the list_keys function using a non existing registry key
    """
    expected = (False, f"Cannot find key: HKLM\\{fake_key}")
    assert win_reg.list_keys(hive="HKLM", key=fake_key) == expected


def test_list_keys_access_denied(fake_key):
    """
    Test the list_keys function using a registry key when access is denied
    """
    expected = (False, f"Access is denied: HKLM\\{fake_key}")
    mock_error = MagicMock(
        side_effect=win32api.error(5, "RegOpenKeyEx", "Access is denied")
    )
    with patch("salt.utils.win_reg.win32api.RegOpenKeyEx", mock_error):
        assert win_reg.list_keys(hive="HKLM", key=fake_key) == expected


def test_list_keys_invalid_hive():
    """
    Test the list_keys function when passing an invalid hive
    """
    pytest.raises(
        CommandExecutionError,
        win_reg.list_keys,
        hive="BADHIVE",
        key="SOFTWARE\\Microsoft",
    )


def test_list_keys_unknown_key_error():
    """
    Tests the list_keys function with an unknown key error
    """
    mock_error = MagicMock(
        side_effect=win32api.error(123, "RegOpenKeyEx", "Unknown error")
    )
    with patch("salt.utils.win_reg.win32api.RegOpenKeyEx", mock_error):
        pytest.raises(
            win32api.error,
            win_reg.list_keys,
            hive="HKLM",
            key="SOFTWARE\\Microsoft",
        )


def test_list_values_existing():
    """
    Test the list_values function using a well known registry key
    """
    values = win_reg.list_values(
        hive="HKLM", key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion"
    )
    keys = []
    for value in values:
        keys.append(value["vname"])
    assert "ProgramFilesDir" in keys


def test_list_values_non_existing(fake_key):
    """
    Test the list_values function using a non existing registry key
    """
    expected = (False, f"Cannot find key: HKLM\\{fake_key}")
    assert win_reg.list_values(hive="HKLM", key=fake_key) == expected


def test_list_values_access_denied(fake_key):
    """
    Test the list_values function using a registry key when access is denied
    """
    expected = (False, f"Access is denied: HKLM\\{fake_key}")
    mock_error = MagicMock(
        side_effect=win32api.error(5, "RegOpenKeyEx", "Access is denied")
    )
    with patch("salt.utils.win_reg.win32api.RegOpenKeyEx", mock_error):
        assert win_reg.list_values(hive="HKLM", key=fake_key) == expected


def test_list_values_invalid_hive():
    """
    Test the list_values function when passing an invalid hive
    """
    pytest.raises(
        CommandExecutionError,
        win_reg.list_values,
        hive="BADHIVE",
        key="SOFTWARE\\Microsoft",
    )


def test_list_values_unknown_key_error():
    """
    Tests the list_values function with an unknown key error
    """
    mock_error = MagicMock(
        side_effect=win32api.error(123, "RegOpenKeyEx", "Unknown error")
    )
    with patch("salt.utils.win_reg.win32api.RegOpenKeyEx", mock_error):
        pytest.raises(
            win32api.error,
            win_reg.list_values,
            hive="HKLM",
            key="SOFTWARE\\Microsoft",
        )


def test_read_value_existing():
    """
    Test the read_value function using a well known registry value
    """
    ret = win_reg.read_value(
        hive="HKLM",
        key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion",
        vname="ProgramFilesPath",
    )
    assert ret["vdata"] == "%ProgramFiles%"


def test_read_value_default():
    """
    Test the read_value function reading the default value using a well
    known registry key
    """
    ret = win_reg.read_value(
        hive="HKLM", key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion"
    )
    assert ret["vdata"] == "(value not set)"


def test_read_value_non_existing():
    """
    Test the read_value function using a non existing value pair
    """
    expected = {
        "comment": (
            "Cannot find fake_name in HKLM\\SOFTWARE\\Microsoft\\"
            "Windows\\CurrentVersion"
        ),
        "vdata": None,
        "vtype": None,
        "vname": "fake_name",
        "success": False,
        "hive": "HKLM",
        "key": "SOFTWARE\\Microsoft\\Windows\\CurrentVersion",
    }
    assert (
        win_reg.read_value(
            hive="HKLM",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion",
            vname="fake_name",
        )
        == expected
    )


def test_read_value_non_existing_key(fake_key):
    """
    Test the read_value function using a non existing registry key
    """
    expected = {
        "comment": f"Cannot find key: HKLM\\{fake_key}",
        "vdata": None,
        "vtype": None,
        "vname": "fake_name",
        "success": False,
        "hive": "HKLM",
        "key": fake_key,
    }
    assert win_reg.read_value(hive="HKLM", key=fake_key, vname="fake_name") == expected


def test_read_value_access_denied(fake_key):
    """
    Test the read_value function using a registry key when access is denied
    """
    expected = {
        "comment": f"Access is denied: HKLM\\{fake_key}",
        "vdata": None,
        "vtype": None,
        "vname": "fake_name",
        "success": False,
        "hive": "HKLM",
        "key": fake_key,
    }
    mock_error = MagicMock(
        side_effect=win32api.error(5, "RegOpenKeyEx", "Access is denied")
    )
    with patch("salt.utils.win_reg.win32api.RegOpenKeyEx", mock_error):
        assert (
            win_reg.read_value(hive="HKLM", key=fake_key, vname="fake_name") == expected
        )


def test_read_value_invalid_hive():
    """
    Test the read_value function when passing an invalid hive
    """
    pytest.raises(
        CommandExecutionError,
        win_reg.read_value,
        hive="BADHIVE",
        key="SOFTWARE\\Microsoft",
        vname="ProgramFilesPath",
    )


def test_read_value_unknown_key_error():
    """
    Tests the read_value function with an unknown key error
    """
    mock_error = MagicMock(
        side_effect=win32api.error(123, "RegOpenKeyEx", "Unknown error")
    )
    with patch("salt.utils.win_reg.win32api.RegOpenKeyEx", mock_error):
        pytest.raises(
            win32api.error,
            win_reg.read_value,
            hive="HKLM",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion",
            vname="ProgramFilesPath",
        )


def test_read_value_unknown_value_error():
    """
    Tests the read_value function with an unknown value error
    """
    mock_error = MagicMock(
        side_effect=win32api.error(123, "RegQueryValueEx", "Unknown error")
    )
    with patch("salt.utils.win_reg.win32api.RegQueryValueEx", mock_error):
        pytest.raises(
            win32api.error,
            win_reg.read_value,
            hive="HKLM",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion",
            vname="ProgramFilesPath",
        )


@pytest.mark.destructive_test
def test_read_value_multi_sz_empty_list(fake_key):
    """
    An empty REG_MULTI_SZ value should return an empty list, not None
    """
    try:
        assert win_reg.set_value(
            hive="HKLM",
            key=fake_key,
            vname="empty_list",
            vdata=[],
            vtype="REG_MULTI_SZ",
        )
        expected = {
            "hive": "HKLM",
            "key": fake_key,
            "success": True,
            "vdata": [],
            "vname": "empty_list",
            "vtype": "REG_MULTI_SZ",
        }
        assert (
            win_reg.read_value(
                hive="HKLM",
                key=fake_key,
                vname="empty_list",
            )
            == expected
        )
    finally:
        win_reg.delete_key_recursive(hive="HKLM", key=fake_key)


@pytest.mark.destructive_test
def test_set_value(fake_key):
    """
    Test the set_value function
    """
    try:
        assert win_reg.set_value(
            hive="HKLM", key=fake_key, vname="fake_name", vdata="fake_data"
        )
        expected = {
            "hive": "HKLM",
            "key": fake_key,
            "success": True,
            "vdata": "fake_data",
            "vname": "fake_name",
            "vtype": "REG_SZ",
        }
        assert (
            win_reg.read_value(hive="HKLM", key=fake_key, vname="fake_name") == expected
        )
    finally:
        win_reg.delete_key_recursive(hive="HKLM", key=fake_key)


@pytest.mark.destructive_test
def test_set_value_default(fake_key):
    """
    Test the set_value function on the default value
    """
    try:
        assert win_reg.set_value(hive="HKLM", key=fake_key, vdata="fake_default_data")
        expected = {
            "hive": "HKLM",
            "key": fake_key,
            "success": True,
            "vdata": "fake_default_data",
            "vname": "(Default)",
            "vtype": "REG_SZ",
        }
        assert win_reg.read_value(hive="HKLM", key=fake_key) == expected
    finally:
        win_reg.delete_key_recursive(hive="HKLM", key=fake_key)


@pytest.mark.destructive_test
def test_set_value_unicode_key(fake_key, unicode_key):
    """
    Test the set_value function on a unicode key
    """
    try:
        assert win_reg.set_value(
            hive="HKLM",
            key="\\".join([fake_key, unicode_key]),
            vname="fake_name",
            vdata="fake_value",
        )
        expected = {
            "hive": "HKLM",
            "key": "\\".join([fake_key, unicode_key]),
            "success": True,
            "vdata": "fake_value",
            "vname": "fake_name",
            "vtype": "REG_SZ",
        }
        assert (
            win_reg.read_value(
                hive="HKLM",
                key="\\".join([fake_key, unicode_key]),
                vname="fake_name",
            )
            == expected
        )
    finally:
        win_reg.delete_key_recursive(hive="HKLM", key=fake_key)


@pytest.mark.destructive_test
def test_set_value_unicode_value(fake_key, unicode_value):
    """
    Test the set_value function on a unicode value
    """
    try:
        assert win_reg.set_value(
            hive="HKLM", key=fake_key, vname="fake_unicode", vdata=unicode_value
        )
        expected = {
            "hive": "HKLM",
            "key": fake_key,
            "success": True,
            "vdata": unicode_value,
            "vname": "fake_unicode",
            "vtype": "REG_SZ",
        }
        assert (
            win_reg.read_value(hive="HKLM", key=fake_key, vname="fake_unicode")
            == expected
        )
    finally:
        win_reg.delete_key_recursive(hive="HKLM", key=fake_key)


@pytest.mark.destructive_test
def test_set_value_reg_dword(fake_key):
    """
    Test the set_value function on a REG_DWORD value
    """
    try:
        assert win_reg.set_value(
            hive="HKLM",
            key=fake_key,
            vname="dword_value",
            vdata=123,
            vtype="REG_DWORD",
        )
        expected = {
            "hive": "HKLM",
            "key": fake_key,
            "success": True,
            "vdata": 123,
            "vname": "dword_value",
            "vtype": "REG_DWORD",
        }
        assert (
            win_reg.read_value(hive="HKLM", key=fake_key, vname="dword_value")
            == expected
        )
    finally:
        win_reg.delete_key_recursive(hive="HKLM", key=fake_key)


@pytest.mark.destructive_test
def test_set_value_reg_qword(fake_key):
    """
    Test the set_value function on a REG_QWORD value
    """
    try:
        assert win_reg.set_value(
            hive="HKLM",
            key=fake_key,
            vname="qword_value",
            vdata=123,
            vtype="REG_QWORD",
        )
        expected = {
            "hive": "HKLM",
            "key": fake_key,
            "success": True,
            "vdata": 123,
            "vname": "qword_value",
            "vtype": "REG_QWORD",
        }
        assert (
            win_reg.read_value(hive="HKLM", key=fake_key, vname="qword_value")
            == expected
        )
    finally:
        win_reg.delete_key_recursive(hive="HKLM", key=fake_key)


def test_set_value_invalid_hive(fake_key):
    """
    Test the set_value function when passing an invalid hive
    """
    pytest.raises(
        CommandExecutionError,
        win_reg.set_value,
        hive="BADHIVE",
        key=fake_key,
        vname="fake_name",
        vdata="fake_data",
    )


def test_set_value_open_create_failure(fake_key):
    """
    Test the set_value function when there is a problem opening/creating
    the key
    """
    mock_error = MagicMock(
        side_effect=win32api.error(123, "RegCreateKeyEx", "Unknown error")
    )
    with patch("salt.utils.win_reg.win32api.RegCreateKeyEx", mock_error):
        assert not win_reg.set_value(
            hive="HKLM", key=fake_key, vname="fake_name", vdata="fake_data"
        )


def test_set_value_type_error(fake_key):
    """
    Test the set_value function when the wrong type of data is passed
    """
    mock_error = MagicMock(side_effect=TypeError("Mocked TypeError"))
    with patch("salt.utils.win_reg.win32api.RegSetValueEx", mock_error):
        assert not win_reg.set_value(
            hive="HKLM", key=fake_key, vname="fake_name", vdata="fake_data"
        )


def test_set_value_system_error(fake_key):
    """
    Test the set_value function when a SystemError occurs while setting the
    value
    """
    mock_error = MagicMock(side_effect=SystemError("Mocked SystemError"))
    with patch("salt.utils.win_reg.win32api.RegSetValueEx", mock_error):
        assert not win_reg.set_value(
            hive="HKLM", key=fake_key, vname="fake_name", vdata="fake_data"
        )


def test_set_value_value_error(fake_key):
    """
    Test the set_value function when a ValueError occurs while setting the
    value
    """
    mock_error = MagicMock(side_effect=ValueError("Mocked ValueError"))
    with patch("salt.utils.win_reg.win32api.RegSetValueEx", mock_error):
        assert not win_reg.set_value(
            hive="HKLM", key=fake_key, vname="fake_name", vdata="fake_data"
        )


def test_cast_vdata_reg_binary():
    """
    Test the cast_vdata function with REG_BINARY
    Should always return binary data
    """
    vdata = salt.utils.stringutils.to_bytes("test data")
    result = win_reg.cast_vdata(vdata=vdata, vtype="REG_BINARY")
    assert isinstance(result, bytes)

    vdata = salt.utils.stringutils.to_str("test data")
    result = win_reg.cast_vdata(vdata=vdata, vtype="REG_BINARY")
    assert isinstance(result, bytes)

    vdata = salt.utils.stringutils.to_unicode("test data")
    result = win_reg.cast_vdata(vdata=vdata, vtype="REG_BINARY")
    assert isinstance(result, bytes)


def test_cast_vdata_reg_dword():
    """
    Test the cast_vdata function with REG_DWORD
    Should always return integer
    """
    vdata = 1
    expected = 1
    result = win_reg.cast_vdata(vdata=vdata, vtype="REG_DWORD")
    assert result == expected

    vdata = "1"
    result = win_reg.cast_vdata(vdata=vdata, vtype="REG_DWORD")
    assert result == expected

    vdata = "0000001"
    result = win_reg.cast_vdata(vdata=vdata, vtype="REG_DWORD")
    assert result == expected


def test_cast_vdata_reg_expand_sz():
    """
    Test the cast_vdata function with REG_EXPAND_SZ
    Should always return unicode
    """
    vdata = salt.utils.stringutils.to_str("test data")
    result = win_reg.cast_vdata(vdata=vdata, vtype="REG_EXPAND_SZ")
    assert isinstance(result, str)

    vdata = salt.utils.stringutils.to_bytes("test data")
    result = win_reg.cast_vdata(vdata=vdata, vtype="REG_EXPAND_SZ")
    assert isinstance(result, str)


def test_cast_vdata_reg_multi_sz():
    """
    Test the cast_vdata function with REG_MULTI_SZ
    Should always return a list of unicode strings
    """
    vdata = [
        salt.utils.stringutils.to_str("test string"),
        salt.utils.stringutils.to_bytes("test bytes"),
    ]
    result = win_reg.cast_vdata(vdata=vdata, vtype="REG_MULTI_SZ")
    assert isinstance(result, list)
    for item in result:
        assert isinstance(item, str)


def test_cast_vdata_reg_qword():
    """
    Test the cast_vdata function with REG_QWORD
    Should always return a long integer
    `int` is `long` is default on Py3
    """
    vdata = 1
    result = win_reg.cast_vdata(vdata=vdata, vtype="REG_QWORD")
    assert isinstance(result, int)

    vdata = "1"
    result = win_reg.cast_vdata(vdata=vdata, vtype="REG_QWORD")
    assert isinstance(result, int)


def test_cast_vdata_reg_sz():
    """
    Test the cast_vdata function with REG_SZ
    Should always return unicode
    """
    vdata = salt.utils.stringutils.to_str("test data")
    result = win_reg.cast_vdata(vdata=vdata, vtype="REG_SZ")
    assert isinstance(result, str)

    vdata = salt.utils.stringutils.to_bytes("test data")
    result = win_reg.cast_vdata(vdata=vdata, vtype="REG_SZ")
    assert isinstance(result, str)

    vdata = None
    result = win_reg.cast_vdata(vdata=vdata, vtype="REG_SZ")
    assert isinstance(result, str)
    assert result == ""


@pytest.mark.destructive_test
def test_delete_value(fake_key):
    """
    Test the delete_value function
    """
    try:
        assert win_reg.set_value(
            hive="HKLM", key=fake_key, vname="fake_name", vdata="fake_data"
        )
        assert win_reg.delete_value(hive="HKLM", key=fake_key, vname="fake_name")
        assert not win_reg.value_exists(hive="HKLM", key=fake_key, vname="fake_name")
    finally:
        win_reg.delete_key_recursive(hive="HKLM", key=fake_key)


def test_delete_value_non_existing(fake_key):
    """
    Test the delete_value function on non existing value
    """
    mock_error = MagicMock(
        side_effect=win32api.error(2, "RegOpenKeyEx", "Unknown error")
    )
    with patch("salt.utils.win_reg.win32api.RegOpenKeyEx", mock_error):
        assert (
            win_reg.delete_value(hive="HKLM", key=fake_key, vname="fake_name") is None
        )


def test_delete_value_invalid_hive(fake_key):
    """
    Test the delete_value function when passing an invalid hive
    """
    pytest.raises(
        CommandExecutionError,
        win_reg.delete_value,
        hive="BADHIVE",
        key=fake_key,
        vname="fake_name",
    )


def test_delete_value_unknown_error(fake_key):
    """
    Test the delete_value function when there is a problem opening the key
    """
    mock_error = MagicMock(
        side_effect=win32api.error(123, "RegOpenKeyEx", "Unknown error")
    )
    with patch("salt.utils.win_reg.win32api.RegOpenKeyEx", mock_error):
        pytest.raises(
            win32api.error,
            win_reg.delete_value,
            hive="HKLM",
            key=fake_key,
            vname="fake_name",
        )


@pytest.mark.destructive_test
def test_delete_value_unicode(fake_key, unicode_value):
    """
    Test the delete_value function on a unicode value
    """
    try:
        assert win_reg.set_value(
            hive="HKLM", key=fake_key, vname="fake_unicode", vdata=unicode_value
        )
        assert win_reg.delete_value(hive="HKLM", key=fake_key, vname="fake_unicode")
        assert not win_reg.value_exists(hive="HKLM", key=fake_key, vname="fake_unicode")
    finally:
        win_reg.delete_key_recursive(hive="HKLM", key=fake_key)


@pytest.mark.destructive_test
def test_delete_value_unicode_vname(fake_key, unicode_key):
    """
    Test the delete_value function on a unicode vname
    """
    try:
        assert win_reg.set_value(
            hive="HKLM", key=fake_key, vname=unicode_key, vdata="junk data"
        )
        assert win_reg.delete_value(hive="HKLM", key=fake_key, vname=unicode_key)
        assert not win_reg.value_exists(hive="HKLM", key=fake_key, vname=unicode_key)
    finally:
        win_reg.delete_key_recursive(hive="HKLM", key=fake_key)


@pytest.mark.destructive_test
def test_delete_value_unicode_key(fake_key, unicode_key):
    """
    Test the delete_value function on a unicode key
    """
    try:
        assert win_reg.set_value(
            hive="HKLM",
            key="\\".join([fake_key, unicode_key]),
            vname="fake_name",
            vdata="junk data",
        )
        assert win_reg.delete_value(
            hive="HKLM", key="\\".join([fake_key, unicode_key]), vname="fake_name"
        )
        assert not win_reg.value_exists(
            hive="HKLM", key="\\".join([fake_key, unicode_key]), vname="fake_name"
        )
    finally:
        win_reg.delete_key_recursive(hive="HKLM", key=fake_key)


def test_delete_key_recursive_invalid_hive(fake_key):
    """
    Test the delete_key_recursive function when passing an invalid hive
    """
    pytest.raises(
        CommandExecutionError,
        win_reg.delete_key_recursive,
        hive="BADHIVE",
        key=fake_key,
    )


def test_delete_key_recursive_key_not_found(fake_key):
    """
    Test the delete_key_recursive function when the passed key to delete is
    not found.
    """
    assert not win_reg.key_exists(hive="HKLM", key=fake_key)
    assert not win_reg.delete_key_recursive(hive="HKLM", key=fake_key)


def test_delete_key_recursive_too_close():
    """
    Test the delete_key_recursive function when the passed key to delete is
    too close to root, such as
    """
    mock_true = MagicMock(return_value=True)
    with patch("salt.utils.win_reg.key_exists", mock_true):
        assert not win_reg.delete_key_recursive(hive="HKLM", key="FAKE_KEY")


@pytest.mark.destructive_test
def test_delete_key_recursive(fake_key):
    """
    Test the delete_key_recursive function
    """
    try:
        assert win_reg.set_value(
            hive="HKLM", key=fake_key, vname="fake_name", vdata="fake_value"
        )
        expected = {"Deleted": ["\\".join(["HKLM", fake_key])], "Failed": []}
        assert win_reg.delete_key_recursive(hive="HKLM", key=fake_key) == expected
        assert not win_reg.key_exists(hive="HKLM", key=fake_key)
    finally:
        win_reg.delete_key_recursive(hive="HKLM", key=fake_key)


@pytest.mark.destructive_test
def test_delete_key_recursive_failed_to_open_key(fake_key):
    """
    Test the delete_key_recursive function on failure to open the key
    """
    try:
        assert win_reg.set_value(
            hive="HKLM", key=fake_key, vname="fake_name", vdata="fake_value"
        )
        expected = {
            "Deleted": [],
            "Failed": ["\\".join(["HKLM", fake_key]) + " Failed to connect to key"],
        }
        mock_true = MagicMock(return_value=True)
        mock_error = MagicMock(
            side_effect=[
                1,
                win32api.error(3, "RegOpenKeyEx", "Failed to connect to key"),
            ]
        )
        with patch("salt.utils.win_reg.key_exists", mock_true), patch(
            "salt.utils.win_reg.win32api.RegOpenKeyEx", mock_error
        ):
            assert win_reg.delete_key_recursive(hive="HKLM", key=fake_key) == expected
    finally:
        win_reg.delete_key_recursive(hive="HKLM", key=fake_key)


@pytest.mark.destructive_test
def test_delete_key_recursive_failed_to_delete(fake_key):
    """
    Test the delete_key_recursive function on failure to delete a key
    """
    try:
        assert win_reg.set_value(
            hive="HKLM", key=fake_key, vname="fake_name", vdata="fake_value"
        )
        expected = {
            "Deleted": [],
            "Failed": ["\\".join(["HKLM", fake_key]) + " Unknown error"],
        }
        # pylint: disable=undefined-variable
        mock_error = MagicMock(side_effect=WindowsError("Unknown error"))
        # pylint: enable=undefined-variable
        with patch("salt.utils.win_reg.win32api.RegDeleteKey", mock_error):
            assert win_reg.delete_key_recursive(hive="HKLM", key=fake_key) == expected
    finally:
        win_reg.delete_key_recursive(hive="HKLM", key=fake_key)


@pytest.mark.destructive_test
def test_delete_key_recursive_unicode(fake_key, unicode_key):
    """
    Test the delete_key_recursive function on value within a unicode key
    """
    try:
        assert win_reg.set_value(
            hive="HKLM",
            key="\\".join([fake_key, unicode_key]),
            vname="fake_name",
            vdata="fake_value",
        )
        expected = {
            "Deleted": ["\\".join(["HKLM", fake_key, unicode_key])],
            "Failed": [],
        }
        assert (
            win_reg.delete_key_recursive(
                hive="HKLM", key="\\".join([fake_key, unicode_key])
            )
            == expected
        )
    finally:
        win_reg.delete_key_recursive(hive="HKLM", key=fake_key)


def test__to_unicode_int():
    """
    Test the ``_to_unicode`` function when it receives an integer value.
    Should return a unicode value, which is str in PY3.
    """
    assert isinstance(win_reg._to_unicode(1), str)



import salt.channel.client
from tests.support.mock import MagicMock, patch


async def test_async_pub_channel_connect_cb(minion_opts):
    """
    Validate connect_callback closes the request channel it creates.
    """
    minion_opts["master_uri"] = "tcp://127.0.0.1:4506"
    minion_opts["master_ip"] = "127.0.0.1"
    with salt.channel.client.AsyncPubChannel.factory(minion_opts) as channel:

        async def send_id(*args):
            return

        channel.send_id = send_id
        channel._reconnected = True

        mock = MagicMock(salt.channel.client.AsyncReqChannel)
        mock.__enter__ = lambda self: mock

        with patch("salt.channel.client.AsyncReqChannel.factory", return_value=mock):
            await channel.connect_callback(None)
            mock.send.assert_called_once()
            mock.__exit__.assert_called_once()



import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
]


def test_symlink(file, tmp_path):
    """
    file.symlink
    """
    symlink = tmp_path / "symlink"
    target = tmp_path / "target"

    # Make sure the symlink target exists
    target.mkdir()

    ret = file.symlink(str(symlink), target=str(target))
    expected = {
        "name": str(symlink),
        "changes": {"new": str(symlink)},
        "result": True,
        "comment": f"Created new symlink {symlink} -> {target}",
    }
    assert ret.filtered == expected
    assert symlink.exists()
    assert symlink.is_symlink()


def test_symlink_test(file, tmp_path):
    """
    file.symlink
    """
    symlink = tmp_path / "symlink"
    target = tmp_path / "target"

    # Make sure the symlink target exists
    target.mkdir()

    ret = file.symlink(str(symlink), target=str(target), test=True)
    assert ret.result is None
    assert symlink.exists() is False



"""
    :codeauthor: Gareth J. Greenaway <ggreenaway@vmware.com>
"""

import logging
import subprocess

import pytest

import salt.modules.beacons as beaconmod
import salt.states.beacon as beaconstate
import salt.states.service as service
import salt.utils.platform
from salt.utils.event import SaltEvent
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


def _check_systemctl():
    if not hasattr(_check_systemctl, "memo"):
        if not salt.utils.platform.is_linux():
            _check_systemctl.memo = False
        else:
            proc = subprocess.run(["systemctl"], capture_output=True, check=False)
            _check_systemctl.memo = (
                b"Failed to get D-Bus connection: No such file or directory"
                in proc.stderr
                or b"Failed to connect to bus: No such file or directory" in proc.stderr
            )
    return _check_systemctl.memo


def func(name):
    """
    Mock func method
    """
    return name


@pytest.fixture
def configure_loader_modules():
    return {
        service: {
            "__env__": "base",
            "__salt__": {},
            "__opts__": {"test": False, "cachedir": ""},
            "__instance_id__": "",
            "__low__": {},
            "__utils__": {},
            "__context__": {},
        },
        beaconstate: {"__salt__": {}, "__opts__": {}},
        beaconmod: {"__salt__": {}, "__opts__": {}},
    }


def test_get_systemd_only():
    """
    Test the _get_system_only function
    """

    def test_func(cats, dogs, no_block):
        pass

    with patch.object(service._get_systemd_only, "HAS_SYSTEMD", True, create=True):
        ret, warnings = service._get_systemd_only(
            test_func, {"cats": 1, "no_block": 2, "unmask": 3}
        )
        assert len(warnings) == 0
        assert ret == {"no_block": 2}

        ret, warnings = service._get_systemd_only(test_func, {"cats": 1, "unmask": 3})

        assert len(warnings) == 0
        assert ret == {}


def test_get_systemd_only_platform():
    """
    Test the _get_system_only function on unsupported platforms
    """

    def test_func(cats, dogs, no_block):
        pass

    with patch.object(service._get_systemd_only, "HAS_SYSTEMD", False, create=True):
        ret, warnings = service._get_systemd_only(
            test_func, {"cats": 1, "no_block": 2, "unmask": 3}
        )

        assert warnings == ["The 'no_block' argument is not supported by this platform"]
        assert ret == {}

        ret, warnings = service._get_systemd_only(test_func, {"cats": 1, "unmask": 3})

        assert len(warnings) == 0
        assert ret == {}


def test_get_systemd_only_no_mock():
    """
    Test the _get_system_only without mocking
    """

    def test_func(cats, dogs, no_block):
        pass

    ret, warnings = service._get_systemd_only(
        test_func, {"cats": 1, "no_block": 2, "unmask": 3}
    )

    assert isinstance(ret, dict)
    assert isinstance(warnings, list)


def test_running():
    """
    Test to verify that the service is running
    """
    ret = [
        {"comment": "", "changes": {}, "name": "salt", "result": True},
        {
            "changes": {},
            "comment": "The service salt is already running",
            "name": "salt",
            "result": True,
        },
        {
            "changes": "saltstack",
            "comment": "The service salt is already running",
            "name": "salt",
            "result": True,
        },
        {
            "changes": {},
            "comment": "Service salt is set to start",
            "name": "salt",
            "result": None,
        },
        {
            "changes": "saltstack",
            "comment": "Started service salt",
            "name": "salt",
            "result": True,
        },
        {
            "changes": {},
            "comment": "The service salt is already running",
            "name": "salt",
            "result": True,
        },
        {
            "changes": "saltstack",
            "comment": "Service salt failed to start",
            "name": "salt",
            "result": False,
        },
        {
            "changes": "saltstack",
            "comment": (
                "Started service salt\nService masking not available on this minion"
            ),
            "name": "salt",
            "result": True,
        },
        {
            "changes": "saltstack",
            "comment": (
                "Started service salt\nService masking not available on this minion"
            ),
            "name": "salt",
            "result": True,
        },
        {
            "changes": {},
            "comment": (
                "The service salt is disabled but enable is not True. Set enable to"
                " True to successfully start the service."
            ),
            "name": "salt",
            "result": False,
        },
        {
            "changes": {},
            "comment": "The service salt is set to restart",
            "name": "salt",
            "result": None,
        },
    ]

    tmock = MagicMock(return_value=True)
    fmock = MagicMock(return_value=False)
    vmock = MagicMock(return_value="salt")
    with patch.object(service, "_enabled_used_error", vmock):
        assert service.running("salt", enabled=1) == "salt"

    with patch.object(service, "_available", fmock):
        assert service.running("salt") == ret[0]

    with patch.object(service, "_available", tmock):
        with patch.dict(service.__opts__, {"test": False}):
            with patch.dict(
                service.__salt__,
                {"service.enabled": tmock, "service.status": tmock},
            ):
                assert service.running("salt") == ret[1]

            mock = MagicMock(return_value={"changes": "saltstack"})
            with patch.dict(
                service.__salt__,
                {
                    "service.enabled": MagicMock(side_effect=[False, True]),
                    "service.status": tmock,
                },
            ):
                with patch.object(service, "_enable", mock):
                    assert service.running("salt", True) == ret[2]

            with patch.dict(
                service.__salt__,
                {
                    "service.enabled": MagicMock(side_effect=[True, False]),
                    "service.status": tmock,
                },
            ):
                with patch.object(service, "_disable", mock):
                    assert service.running("salt", False) == ret[2]

            with patch.dict(
                service.__salt__,
                {
                    "service.status": MagicMock(side_effect=[False, True]),
                    "service.enabled": MagicMock(side_effect=[False, True]),
                    "service.start": MagicMock(return_value="stack"),
                },
            ):
                with patch.object(
                    service,
                    "_enable",
                    MagicMock(return_value={"changes": "saltstack"}),
                ):
                    assert service.running("salt", True) == ret[4]

            with patch.dict(
                service.__salt__,
                {
                    "service.status": MagicMock(side_effect=[False, True]),
                    "service.enabled": MagicMock(side_effect=[False, True]),
                    "service.unmask": MagicMock(side_effect=[False, True]),
                    "service.start": MagicMock(return_value="stack"),
                },
            ):
                with patch.object(
                    service,
                    "_enable",
                    MagicMock(return_value={"changes": "saltstack"}),
                ):
                    assert service.running("salt", True, unmask=True) == ret[7]

        with patch.dict(service.__opts__, {"test": True}):
            with patch.dict(service.__salt__, {"service.status": tmock}):
                assert service.running("salt") == ret[5]

            with patch.dict(service.__salt__, {"service.status": fmock}):
                assert service.running("salt") == ret[3]

        with patch.dict(service.__opts__, {"test": False}):
            with patch.dict(
                service.__salt__,
                {
                    "service.status": MagicMock(side_effect=[False, False]),
                    "service.enabled": MagicMock(side_effect=[True, True]),
                    "service.start": MagicMock(return_value="stack"),
                },
            ):
                with patch.object(
                    service,
                    "_enable",
                    MagicMock(return_value={"changes": "saltstack"}),
                ):
                    assert service.running("salt", True) == ret[6]
            # test some unique cases simulating Windows
            with patch.object(salt.utils.platform, "is_windows", tmock):
                # We should fail if a service is disabled on Windows and enable
                # isn't set.
                with patch.dict(
                    service.__salt__,
                    {
                        "service.status": fmock,
                        "service.enabled": fmock,
                        "service.start": tmock,
                    },
                ):
                    assert service.running("salt", None) == ret[9]
                    assert service.__context__ == {"service.state": "running"}
            # test some unique cases simulating macOS
            with patch.object(salt.utils.platform, "is_darwin", tmock):
                # We should fail if a service is disabled on macOS and enable
                # isn't set.
                with patch.dict(
                    service.__salt__,
                    {
                        "service.status": fmock,
                        "service.enabled": fmock,
                        "service.start": tmock,
                    },
                ):
                    assert service.running("salt", None) == ret[9]
                    assert service.__context__ == {"service.state": "running"}
                # test enabling a service prior starting it on macOS
                with patch.dict(
                    service.__salt__,
                    {
                        "service.status": MagicMock(side_effect=[False, "loaded"]),
                        "service.enabled": MagicMock(side_effect=[False, True]),
                        "service.start": tmock,
                    },
                ):
                    with patch.object(
                        service,
                        "_enable",
                        MagicMock(return_value={"changes": "saltstack"}),
                    ):
                        assert service.running("salt", True) == ret[4]
                        assert service.__context__ == {"service.state": "running"}
                # if an enable attempt fails on macOS or windows then a
                # disabled service will always fail to start.
                with patch.dict(
                    service.__salt__,
                    {
                        "service.status": fmock,
                        "service.enabled": fmock,
                        "service.start": fmock,
                    },
                ):
                    with patch.object(
                        service,
                        "_enable",
                        MagicMock(
                            return_value={"changes": "saltstack", "result": False}
                        ),
                    ):
                        assert service.running("salt", True) == ret[6]
                        assert service.__context__ == {"service.state": "running"}


def test_running_in_offline_mode():
    """
    Tests the case in which a service.running state is executed on an offline environemnt

    """
    name = "thisisnotarealservice"
    with patch.object(service, "_offline", MagicMock(return_value=True)):
        ret = service.running(name=name)
        assert ret == {
            "changes": {},
            "comment": "Running in OFFLINE mode. Nothing to do",
            "result": True,
            "name": name,
        }


def test_dead():
    """
    Test to ensure that the named service is dead
    """
    ret = [
        {"changes": {}, "comment": "", "name": "salt", "result": True},
        {
            "changes": "saltstack",
            "comment": "The service salt is already dead",
            "name": "salt",
            "result": True,
        },
        {
            "changes": {},
            "comment": "Service salt is set to be killed",
            "name": "salt",
            "result": None,
        },
        {
            "changes": "saltstack",
            "comment": "Service salt was killed",
            "name": "salt",
            "result": True,
        },
        {
            "changes": {},
            "comment": "Service salt failed to die",
            "name": "salt",
            "result": False,
        },
        {
            "changes": "saltstack",
            "comment": "The service salt is already dead",
            "name": "salt",
            "result": True,
        },
    ]
    info_mock = MagicMock(return_value={"StartType": ""})

    mock = MagicMock(return_value="salt")
    with patch.object(service, "_enabled_used_error", mock):
        assert service.dead("salt", enabled=1) == "salt"

    tmock = MagicMock(return_value=True)
    fmock = MagicMock(return_value=False)
    with patch.object(service, "_available", fmock):
        assert service.dead("salt") == ret[0]

    with patch.object(service, "_available", tmock):
        mock = MagicMock(return_value={"changes": "saltstack"})
        with patch.dict(service.__opts__, {"test": True}):
            with patch.dict(
                service.__salt__,
                {
                    "service.enabled": fmock,
                    "service.stop": tmock,
                    "service.status": fmock,
                    "service.info": info_mock,
                },
            ):
                with patch.object(service, "_enable", mock):
                    assert service.dead("salt", True) == ret[5]

            with patch.dict(
                service.__salt__,
                {
                    "service.enabled": tmock,
                    "service.status": tmock,
                    "service.info": info_mock,
                },
            ):
                assert service.dead("salt") == ret[2]

        with patch.dict(service.__opts__, {"test": False}):
            with patch.dict(
                service.__salt__,
                {
                    "service.enabled": fmock,
                    "service.stop": tmock,
                    "service.status": fmock,
                    "service.info": info_mock,
                },
            ):
                with patch.object(service, "_enable", mock):
                    assert service.dead("salt", True) == ret[1]

            with patch.dict(
                service.__salt__,
                {
                    "service.enabled": MagicMock(side_effect=[True, True, False]),
                    "service.status": MagicMock(side_effect=[True, False, False]),
                    "service.stop": MagicMock(return_value="stack"),
                    "service.info": info_mock,
                },
            ):
                with patch.object(
                    service,
                    "_enable",
                    MagicMock(return_value={"changes": "saltstack"}),
                ):
                    assert service.dead("salt", True) == ret[3]

            # test an initd which a wrong status (True even if dead)
            with patch.dict(
                service.__salt__,
                {
                    "service.enabled": MagicMock(side_effect=[False, False, False]),
                    "service.status": MagicMock(side_effect=[True, True, True]),
                    "service.stop": MagicMock(return_value="stack"),
                    "service.info": info_mock,
                },
            ):
                with patch.object(service, "_disable", MagicMock(return_value={})):
                    assert service.dead("salt", False) == ret[4]

        assert service.__context__ == {"service.state": "dead"}


def test_dead_with_missing_service():
    """
    Tests the case in which a service.dead state is executed on a state
    which does not exist.

    See https://github.com/saltstack/salt/issues/37511
    """
    name = "thisisnotarealservice"
    with patch.dict(
        service.__salt__, {"service.available": MagicMock(return_value=False)}
    ):
        ret = service.dead(name=name)
        assert ret == {
            "changes": {},
            "comment": f"The named service {name} is not available",
            "result": True,
            "name": name,
        }


def test_dead_in_offline_mode():
    """
    Tests the case in which a service.dead state is executed on an offline environemnt

    """
    name = "thisisnotarealservice"
    with patch.object(service, "_offline", MagicMock(return_value=True)):
        ret = service.dead(name=name)
        assert ret == {
            "changes": {},
            "comment": "Running in OFFLINE mode. Nothing to do",
            "result": True,
            "name": name,
        }


def test_enabled():
    """
    Test to verify that the service is enabled
    """
    ret = {"changes": "saltstack", "comment": "", "name": "salt", "result": True}
    mock = MagicMock(return_value={"changes": "saltstack"})
    with patch.object(service, "_enable", mock):
        assert service.enabled("salt") == ret
        assert service.__context__ == {"service.state": "enabled"}


def test_enabled_in_test_mode():
    ret = {
        "changes": {},
        "comment": "Service salt not present; if created in this state run, it would have been enabled",
        "name": "salt",
        "result": None,
    }
    mock = MagicMock(
        return_value={
            "result": "False",
            "comment": "The named service salt is not available",
        }
    )
    with patch.object(service, "_enable", mock), patch.dict(
        service.__opts__, {"test": True}
    ):
        assert service.enabled("salt") == ret
        assert service.__context__ == {"service.state": "enabled"}


def test_disabled():
    """
    Test to verify that the service is disabled
    """
    ret = {"changes": "saltstack", "comment": "", "name": "salt", "result": True}
    mock = MagicMock(return_value={"changes": "saltstack"})
    with patch.object(service, "_disable", mock):
        assert service.disabled("salt") == ret
        assert service.__context__ == {"service.state": "disabled"}


def test_mod_watch():
    """
    Test to the service watcher, called to invoke the watch command.
    """
    ret = [
        {
            "changes": {},
            "comment": "Service is already stopped",
            "name": "salt",
            "result": True,
        },
        {
            "changes": {},
            "comment": "Unable to trigger watch for service.stack",
            "name": "salt",
            "result": False,
        },
        {
            "changes": {},
            "comment": "Service is set to be started",
            "name": "salt",
            "result": None,
        },
        {
            "changes": {"salt": "salt"},
            "comment": "Service started",
            "name": "salt",
            "result": "salt",
        },
    ]

    mock = MagicMock(return_value=False)
    with patch.dict(service.__salt__, {"service.status": mock}):
        assert service.mod_watch("salt", "dead") == ret[0]

        with patch.dict(service.__salt__, {"service.start": func}):
            with patch.dict(service.__opts__, {"test": True}):
                assert service.mod_watch("salt", "running") == ret[2]

            with patch.dict(service.__opts__, {"test": False}):
                assert service.mod_watch("salt", "running") == ret[3]

        assert service.mod_watch("salt", "stack") == ret[1]


def test_mod_beacon(tmp_path):
    """
    Test to create a beacon based on a service
    """
    name = "sshd"

    with patch.dict(service.__salt__, {"beacons.list": MagicMock(return_value={})}):
        with patch.dict(service.__states__, {"beacon.present": beaconstate.present}):
            ret = service.mod_beacon(name, sfun="copy")
            expected = {
                "name": name,
                "changes": {},
                "result": False,
                "comment": "service.copy does not work with the beacon state function",
            }

            assert ret == expected

    event_returns = [
        {
            "complete": True,
            "tag": "/salt/minion/minion_beacons_list_complete",
            "beacons": {},
        },
        {
            "complete": True,
            "tag": "/salt/minion/minion_beacons_list_complete",
            "beacons": {},
        },
        {
            "complete": True,
            "tag": "/salt/minion/minion_beacons_list_available_complete",
            "beacons": ["service"],
        },
        {
            "valid": True,
            "tag": "/salt/minion/minion_beacon_validation_complete",
            "vcomment": "Valid beacon configuration",
        },
        {
            "complete": True,
            "tag": "/salt/minion/minion_beacon_add_complete",
            "beacons": {
                "beacon_service_sshd": [
                    {
                        "services": {
                            "sshd": {
                                "onchangeonly": True,
                                "delay": 0,
                                "uncleanshutdown": None,
                                "emitatstartup": False,
                            },
                        }
                    },
                    {"interval": 60},
                    {"beacon_module": "service"},
                ]
            },
        },
    ]
    mock = MagicMock(return_value=True)
    beacon_state_mocks = {
        "beacons.list": beaconmod.list_,
        "beacons.add": beaconmod.add,
        "beacons.list_available": beaconmod.list_available,
        "event.fire": mock,
    }

    beacon_mod_mocks = {"event.fire": mock}

    sock_dir = str(tmp_path / "test-socks")
    with patch.dict(service.__states__, {"beacon.present": beaconstate.present}):
        with patch.dict(beaconstate.__salt__, beacon_state_mocks):
            with patch.dict(beaconmod.__salt__, beacon_mod_mocks):
                with patch.dict(
                    beaconmod.__opts__, {"beacons": {}, "sock_dir": sock_dir}
                ):
                    with patch.object(
                        SaltEvent, "get_event", side_effect=event_returns
                    ):
                        ret = service.mod_beacon(name, sfun="running", beacon="True")
                        expected = {
                            "name": "beacon_service_sshd",
                            "changes": {},
                            "result": True,
                            "comment": "Adding beacon_service_sshd to beacons",
                        }

                        assert ret == expected


@pytest.mark.skipif(_check_systemctl(), reason="systemctl is in a degraded state")
@pytest.mark.skip_on_darwin(reason="service.running is currently failing on OSX")
@pytest.mark.skip_if_not_root
@pytest.mark.destructive_test
@pytest.mark.slow_test
def test_running_with_reload(minion_opts):
    """
    Test that a running service is properly reloaded
    """
    # TODO: This is not a unit test, it interacts with the system. Move to functional.
    minion_opts["grains"] = salt.loader.grains(minion_opts)
    utils = salt.loader.utils(minion_opts)
    modules = salt.loader.minion_mods(minion_opts, utils=utils)

    service_name = "cron"
    cmd_name = "crontab"
    os_family = minion_opts["grains"]["os_family"]
    os_release = minion_opts["grains"]["osrelease"]
    if os_family == "RedHat":
        service_name = "crond"
    elif os_family == "Arch":
        service_name = "sshd"
        cmd_name = "systemctl"
    elif os_family == "MacOS":
        service_name = "org.ntp.ntpd"
        if int(os_release.split(".")[1]) >= 13:
            service_name = "com.openssh.sshd"
    elif os_family == "Windows":
        service_name = "Spooler"

    if os_family != "Windows" and salt.utils.path.which(cmd_name) is None:
        pytest.skip(f"{cmd_name} is not installed")

    pre_srv_enabled = (
        True if service_name in modules["service.get_enabled"]() else False
    )
    post_srv_disable = False
    if not pre_srv_enabled:
        modules["service.enable"](service_name)
        post_srv_disable = True

    try:
        with patch.dict(service.__grains__, minion_opts["grains"]), patch.dict(
            service.__opts__, minion_opts
        ), patch.dict(service.__salt__, modules), patch.dict(
            service.__utils__, utils
        ), patch.dict(
            service.__opts__, {"test": False}
        ), patch(
            "salt.utils.systemd.offline", MagicMock(return_value=False)
        ):
            service.dead(service_name, enable=False)
            result = service.running(name=service_name, enable=True, reload=False)

        if salt.utils.platform.is_windows():
            comment = f"Started service {service_name}"
        else:
            comment = f"Service {service_name} has been enabled, and is running"
        expected = {
            "changes": {service_name: True},
            "comment": comment,
            "name": service_name,
            "result": True,
        }
        assert result == expected
    finally:
        if post_srv_disable:
            modules["service.disable"](service_name)



"""
Test generic roster matching utility.
"""

import os

import pytest

import salt.config
import salt.loader
import salt.utils.roster_matcher
from tests.support import mixins
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase

EXPECTED = {
    "host1": {
        "host": "host1",
        "passwd": "test123",
        "minion_opts": {
            "escape_pods": 2,
            "halon_system_timeout": 30,
            "self_destruct_countdown": 60,
            "some_server": "foo.southeast.example.com",
        },
    },
    "host2": {
        "host": "host2",
        "passwd": "test123",
        "minion_opts": {
            "escape_pods": 2,
            "halon_system_timeout": 30,
            "self_destruct_countdown": 60,
            "some_server": "foo.southeast.example.com",
        },
    },
    "host3": {"host": "host3", "passwd": "test123", "minion_opts": {}},
    "host4": "host4.example.com",  # For testing get_data -> string_types branch
    "host5": None,  # For testing get_data -> False
}


class RosterMatcherTestCase(TestCase, mixins.LoaderModuleMockMixin):
    """
    Test the RosterMatcher Utility
    """

    def setup_loader_modules(self):
        opts = salt.config.master_config(
            os.path.join(RUNTIME_VARS.TMP_CONF_DIR, "master")
        )
        utils = salt.loader.utils(opts, whitelist=["json", "stringutils"])
        runner = salt.loader.runner(opts, utils=utils, whitelist=["salt"])
        return {
            salt.utils.roster_matcher: {
                "__utils__": utils,
                "__opts__": {
                    "ssh_list_nodegroups": {
                        "list_nodegroup": ["host5", "host1", "host2"],
                        "string_nodegroup": "host3,host5,host1,host4",
                    }
                },
                "__runner__": runner,
            }
        }

    def test_get_data(self):
        """
        Test that the get_data method returns the expected dictionaries.
        """
        # We don't care about tgt and tgt_type here.
        roster_matcher = salt.utils.roster_matcher.RosterMatcher(
            EXPECTED, "tgt", "tgt_type"
        )
        self.assertEqual(EXPECTED["host1"], roster_matcher.get_data("host1"))
        self.assertEqual(EXPECTED["host2"], roster_matcher.get_data("host2"))
        self.assertEqual(EXPECTED["host3"], roster_matcher.get_data("host3"))
        self.assertEqual({"host": EXPECTED["host4"]}, roster_matcher.get_data("host4"))

    def test_ret_glob_minions(self):
        """
        Test that we return minions matching a glob.
        """
        result = salt.utils.roster_matcher.targets(EXPECTED, "*[245]", "glob")
        self.assertNotIn("host1", result)
        self.assertIn("host2", result)
        self.assertNotIn("host3", result)
        self.assertIn("host4", result)
        self.assertNotIn("host5", result)

    def test_ret_pcre_minions(self):
        """
        Test that we return minions matching a regular expression.
        """
        result = salt.utils.roster_matcher.targets(EXPECTED, ".*[^23]$", "pcre")
        self.assertIn("host1", result)
        self.assertNotIn("host2", result)
        self.assertNotIn("host3", result)
        self.assertIn("host4", result)
        self.assertNotIn("host5", result)

    def test_ret_literal_list_minions(self):
        """
        Test that we return minions that are in a literal list.
        """
        result = salt.utils.roster_matcher.targets(
            EXPECTED, ["host1", "host2", "host5"], "list"
        )
        self.assertIn("host1", result)
        self.assertIn("host2", result)
        self.assertNotIn("host3", result)
        self.assertNotIn("host4", result)
        self.assertNotIn("host5", result)

    def test_ret_comma_delimited_string_minions(self):
        """
        Test that we return minions that are in a comma-delimited
        string of literal minion names.
        """
        result = salt.utils.roster_matcher.targets(
            EXPECTED, "host5,host3,host2", "list"
        )
        self.assertNotIn("host1", result)
        self.assertIn("host2", result)
        self.assertIn("host3", result)
        self.assertNotIn("host4", result)
        self.assertNotIn("host5", result)

    def test_ret_oops_minions(self):
        """
        Test that we return no minions when we try to use a matching
        method that is not defined.
        """
        result = salt.utils.roster_matcher.targets(EXPECTED, None, "xyzzy")
        self.assertEqual({}, result)

    def test_ret_literal_list_nodegroup_minions(self):
        """
        Test that we return minions that are in a nodegroup
        where the nodegroup expresses a literal list of minion names.
        """
        result = salt.utils.roster_matcher.targets(
            EXPECTED, "list_nodegroup", "nodegroup"
        )
        self.assertIn("host1", result)
        self.assertIn("host2", result)
        self.assertNotIn("host3", result)
        self.assertNotIn("host4", result)
        self.assertNotIn("host5", result)

    def test_ret_comma_delimited_string_nodegroup_minions(self):
        """
        Test that we return minions that are in a nodegroup
        where the nodegroup expresses a comma delimited string
        of minion names.
        """
        result = salt.utils.roster_matcher.targets(
            EXPECTED, "string_nodegroup", "nodegroup"
        )
        self.assertIn("host1", result)
        self.assertNotIn("host2", result)
        self.assertIn("host3", result)
        self.assertIn("host4", result)
        self.assertNotIn("host5", result)

    def test_ret_no_range_installed_minions(self):
        """
        Test that range matcher raises a Runtime Error if seco.range is not installed.
        """
        salt.utils.roster_matcher.HAS_RANGE = False
        with self.assertRaises(RuntimeError):
            salt.utils.roster_matcher.targets(EXPECTED, None, "range")

    @pytest.mark.skipif(
        not salt.utils.roster_matcher.HAS_RANGE, reason="seco.range is not installed"
    )
    def test_ret_range_minions(self):
        """
        Test that range matcher raises a Runtime Error if seco.range is not installed.
        """
        self.fail("Not implemented")



"""
    tests.pytests.unit.beacons.test_log_beacon
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    log beacon test cases
"""

import pytest

import salt.beacons.log_beacon as log_beacon
from tests.support.mock import mock_open, patch


@pytest.fixture
def configure_loader_modules():
    return {log_beacon: {"__context__": {"log.loc": 2}, "__salt__": {}}}


@pytest.fixture
def stub_log_entry():
    return (
        "Jun 29 12:58:51 hostname sshd[6536]: pam_unix(sshd:session): session opened"
        " for user username by (uid=0)\n"
    )


def test_non_list_config():
    config = {}

    ret = log_beacon.validate(config)
    assert ret == (False, "Configuration for log beacon must be a list.")


def test_empty_config():
    config = [{}]

    ret = log_beacon.validate(config)
    assert ret == (False, "Configuration for log beacon must contain file option.")


def test_log_match(stub_log_entry):
    with patch("salt.utils.files.fopen", mock_open(read_data=stub_log_entry)):
        config = [
            {"file": "/var/log/auth.log", "tags": {"sshd": {"regex": ".*sshd.*"}}}
        ]

        ret = log_beacon.validate(config)
        assert ret == (True, "Valid beacon configuration")

        _expected_return = [
            {
                "error": "",
                "match": "yes",
                "raw": stub_log_entry.rstrip("\n"),
                "tag": "sshd",
            }
        ]
        ret = log_beacon.beacon(config)
        assert ret == _expected_return



"""
unit tests for the beacon_module parameter
"""

import logging

import salt.beacons
from tests.support.mock import MagicMock, call, patch

log = logging.getLogger(__name__)


def test_beacon_process(minion_opts):
    """
    Test the process function in the beacon class
    returns the correct information when an exception
    occurs
    """
    minion_opts["id"] = "minion"
    minion_opts["__role"] = "minion"
    minion_opts["beacons"] = {
        "watch_apache": [
            {"processes": {"apache2": "stopped"}},
            {"beacon_module": "ps"},
        ]
    }
    beacon_mock = MagicMock(side_effect=Exception("Global Thermonuclear War"))
    beacon_mock.__globals__ = {}

    beacon = salt.beacons.Beacon(minion_opts, [])

    found = "ps.beacon" in beacon.beacons
    beacon.beacons["ps.beacon"] = beacon_mock
    ret = beacon.process(minion_opts["beacons"], minion_opts["grains"])

    _expected = [
        {
            "tag": "salt/beacon/minion/watch_apache/",
            "error": "Global Thermonuclear War",
            "data": {},
            "beacon_name": "ps",
        }
    ]
    assert ret == _expected


def test_beacon_process_invalid(minion_opts):
    """
    Test the process function in the beacon class
    when the configuration is invalid.
    """
    minion_opts["id"] = "minion"
    minion_opts["__role"] = "minion"

    minion_opts["beacons"] = {"status": {}}

    beacon = salt.beacons.Beacon(minion_opts, [])

    with patch.object(salt.beacons, "log") as log_mock, patch.object(
        salt.beacons.log, "error"
    ) as log_error_mock:
        ret = beacon.process(minion_opts["beacons"], minion_opts["grains"])
        log_error_mock.assert_called_with(
            "Beacon %s configuration invalid, not running.\n%s",
            "status",
            "Configuration for status beacon must be a list.",
        )

    minion_opts["beacons"] = {"mybeacon": {}}

    beacon = salt.beacons.Beacon(minion_opts, [])

    with patch.object(salt.beacons.log, "warning") as log_warn_mock, patch.object(
        salt.beacons.log, "error"
    ) as log_error_mock:
        ret = beacon.process(minion_opts["beacons"], minion_opts["grains"])
        log_warn_mock.assert_called_with(
            "No validate function found for %s, running basic beacon validation.",
            "mybeacon",
        )
        log_error_mock.assert_called_with("Configuration for beacon must be a list.")


def test_beacon_module(minion_opts):
    """
    Test that beacon_module parameter for beacon configuration
    """
    minion_opts["id"] = "minion"
    minion_opts["__role"] = "minion"
    minion_opts["beacons"] = {
        "watch_apache": [
            {"processes": {"apache2": "stopped"}},
            {"beacon_module": "ps"},
        ]
    }
    beacon = salt.beacons.Beacon(minion_opts, [])
    ret = beacon.process(minion_opts["beacons"], minion_opts["grains"])

    _expected = [
        {
            "tag": "salt/beacon/minion/watch_apache/",
            "data": {"id": "minion", "apache2": "Stopped"},
            "beacon_name": "ps",
        }
    ]
    assert ret == _expected

    # Ensure that "beacon_name" is available in the call to the beacon function
    name = "ps.beacon"
    mocked = {name: MagicMock(return_value=_expected)}
    mocked[name].__globals__ = {}
    calls = [
        call(
            [
                {"processes": {"apache2": "stopped"}},
                {"beacon_module": "ps"},
                {"_beacon_name": "watch_apache"},
            ]
        )
    ]
    with patch.object(beacon, "beacons", mocked) as patched:
        beacon.process(minion_opts["beacons"], minion_opts["grains"])
        patched[name].assert_has_calls(calls)



"""
Module for gathering and managing information about MooseFS
"""

import salt.utils.path


def __virtual__():
    """
    Only load if the mfs commands are installed
    """
    if salt.utils.path.which("mfsgetgoal"):
        return "moosefs"
    return (
        False,
        "The moosefs execution module cannot be loaded: the mfsgetgoal binary is not in"
        " the path.",
    )


def dirinfo(path, opts=None):
    """
    Return information on a directory located on the Moose

    CLI Example:

    .. code-block:: bash

        salt '*' moosefs.dirinfo /path/to/dir/ [-[n][h|H]]
    """
    cmd = "mfsdirinfo"
    ret = {}
    if opts:
        cmd += " -" + opts
    cmd += " " + path
    out = __salt__["cmd.run_all"](cmd, python_shell=False)

    output = out["stdout"].splitlines()
    for line in output:
        if not line:
            continue
        comps = line.split(":")
        ret[comps[0].strip()] = comps[1].strip()
    return ret


def fileinfo(path):
    """
    Return information on a file located on the Moose

    CLI Example:

    .. code-block:: bash

        salt '*' moosefs.fileinfo /path/to/dir/
    """
    cmd = "mfsfileinfo " + path
    ret = {}
    chunknum = ""
    out = __salt__["cmd.run_all"](cmd, python_shell=False)

    output = out["stdout"].splitlines()
    for line in output:
        if not line:
            continue
        if "/" in line:
            comps = line.split("/")

            chunknum = comps[0].strip().split(":")
            meta = comps[1].strip().split(" ")

            chunk = chunknum[0].replace("chunk ", "")
            loc = chunknum[1].strip()
            id_ = meta[0].replace("(id:", "")
            ver = meta[1].replace(")", "").replace("ver:", "")

            ret[chunknum[0]] = {
                "chunk": chunk,
                "loc": loc,
                "id": id_,
                "ver": ver,
            }
        if "copy" in line:
            copyinfo = line.strip().split(":")
            ret[chunknum[0]][copyinfo[0]] = {
                "copy": copyinfo[0].replace("copy ", ""),
                "ip": copyinfo[1].strip(),
                "port": copyinfo[2],
            }
    return ret


def mounts():
    """
    Return a list of current MooseFS mounts

    CLI Example:

    .. code-block:: bash

        salt '*' moosefs.mounts
    """
    cmd = "mount"
    ret = {}
    out = __salt__["cmd.run_all"](cmd)

    output = out["stdout"].splitlines()
    for line in output:
        if not line:
            continue
        if "fuse.mfs" in line:
            comps = line.split(" ")
            info1 = comps[0].split(":")
            info2 = info1[1].split("/")
            ret[comps[2]] = {
                "remote": {
                    "master": info1[0],
                    "port": info2[0],
                    "subfolder": "/" + info2[1],
                },
                "local": comps[2],
                "options": (comps[5].replace("(", "").replace(")", "").split(",")),
            }
    return ret


def getgoal(path, opts=None):
    """
    Return goal(s) for a file or directory

    CLI Example:

    .. code-block:: bash

        salt '*' moosefs.getgoal /path/to/file [-[n][h|H]]
        salt '*' moosefs.getgoal /path/to/dir/ [-[n][h|H][r]]
    """
    cmd = "mfsgetgoal"
    ret = {}
    if opts:
        cmd += " -" + opts
    else:
        opts = ""
    cmd += " " + path
    out = __salt__["cmd.run_all"](cmd, python_shell=False)

    output = out["stdout"].splitlines()
    if "r" not in opts:
        goal = output[0].split(": ")
        ret = {
            "goal": goal[1],
        }
    else:
        for line in output:
            if not line:
                continue
            if path in line:
                continue
            comps = line.split()
            keytext = comps[0] + " with goal"
            if keytext not in ret:
                ret[keytext] = {}
            ret[keytext][comps[3]] = comps[5]
    return ret



"""
Riak Salt Module
"""

import salt.utils.path


def __virtual__():
    """
    Only available on systems with Riak installed.
    """
    if salt.utils.path.which("riak"):
        return True
    return (
        False,
        "The riak execution module failed to load: the riak binary is not in the path.",
    )


def __execute_cmd(name, cmd):
    """
    Execute Riak commands
    """
    return __salt__["cmd.run_all"](f"{salt.utils.path.which(name)} {cmd}")


def start():
    """
    Start Riak

    CLI Example:

    .. code-block:: bash

        salt '*' riak.start
    """
    ret = {"comment": "", "success": False}

    cmd = __execute_cmd("riak", "start")

    if cmd["retcode"] != 0:
        ret["comment"] = cmd["stderr"]
    else:
        ret["comment"] = cmd["stdout"]
        ret["success"] = True

    return ret


def stop():
    """
    Stop Riak

    .. versionchanged:: 2015.8.0

    CLI Example:

    .. code-block:: bash

        salt '*' riak.stop
    """
    ret = {"comment": "", "success": False}

    cmd = __execute_cmd("riak", "stop")

    if cmd["retcode"] != 0:
        ret["comment"] = cmd["stderr"]
    else:
        ret["comment"] = cmd["stdout"]
        ret["success"] = True

    return ret


def cluster_join(username, hostname):
    """
    Join a Riak cluster

    .. versionchanged:: 2015.8.0

    CLI Example:

    .. code-block:: bash

        salt '*' riak.cluster_join <user> <host>

    username - The riak username to join the cluster
    hostname - The riak hostname you are connecting to
    """
    ret = {"comment": "", "success": False}

    cmd = __execute_cmd("riak-admin", f"cluster join {username}@{hostname}")

    if cmd["retcode"] != 0:
        ret["comment"] = cmd["stdout"]
    else:
        ret["comment"] = cmd["stdout"]
        ret["success"] = True

    return ret


def cluster_leave(username, hostname):
    """
    Leave a Riak cluster

    .. versionadded:: 2015.8.0

    CLI Example:

    .. code-block:: bash

        salt '*' riak.cluster_leave <username> <host>

    username - The riak username to join the cluster
    hostname - The riak hostname you are connecting to
    """
    ret = {"comment": "", "success": False}

    cmd = __execute_cmd("riak-admin", f"cluster leave {username}@{hostname}")

    if cmd["retcode"] != 0:
        ret["comment"] = cmd["stdout"]
    else:
        ret["comment"] = cmd["stdout"]
        ret["success"] = True

    return ret


def cluster_plan():
    """
    Review Cluster Plan

    .. versionchanged:: 2015.8.0

    CLI Example:

    .. code-block:: bash

        salt '*' riak.cluster_plan
    """
    cmd = __execute_cmd("riak-admin", "cluster plan")

    if cmd["retcode"] != 0:
        return False

    return True


def cluster_commit():
    """
    Commit Cluster Changes

    .. versionchanged:: 2015.8.0

    CLI Example:

    .. code-block:: bash

        salt '*' riak.cluster_commit
    """
    ret = {"comment": "", "success": False}

    cmd = __execute_cmd("riak-admin", "cluster commit")

    if cmd["retcode"] != 0:
        ret["comment"] = cmd["stdout"]
    else:
        ret["comment"] = cmd["stdout"]
        ret["success"] = True

    return ret


def member_status():
    """
    Get cluster member status

    .. versionchanged:: 2015.8.0

    CLI Example:

    .. code-block:: bash

        salt '*' riak.member_status
    """
    ret = {
        "membership": {},
        "summary": {"Valid": 0, "Leaving": 0, "Exiting": 0, "Joining": 0, "Down": 0},
    }

    out = __execute_cmd("riak-admin", "member-status")["stdout"].splitlines()

    for line in out:
        if line.startswith(("=", "-", "Status")):
            continue
        if "/" in line:
            # We're in the summary line
            for item in line.split("/"):
                key, val = item.split(":")
                ret["summary"][key.strip()] = val.strip()

        if len(line.split()) == 4:
            # We're on a node status line
            (status, ring, pending, node) = line.split()

            ret["membership"][node] = {
                "Status": status,
                "Ring": ring,
                "Pending": pending,
            }

    return ret


def status():
    """
    Current node status

    .. versionadded:: 2015.8.0

    CLI Example:

    .. code-block:: bash

        salt '*' riak.status
    """
    ret = {}

    cmd = __execute_cmd("riak-admin", "status")

    for i in cmd["stdout"].splitlines():
        if ":" in i:
            (name, val) = i.split(":", 1)
            ret[name.strip()] = val.strip()

    return ret


def test():
    """
    Runs a test of a few standard Riak operations

    .. versionadded:: 2015.8.0

    CLI Example:

    .. code-block:: bash

        salt '*' riak.test
    """
    ret = {"comment": "", "success": False}

    cmd = __execute_cmd("riak-admin", "test")

    if cmd["retcode"] != 0:
        ret["comment"] = cmd["stdout"]
    else:
        ret["comment"] = cmd["stdout"]
        ret["success"] = True

    return ret


def services():
    """
    List available services on a node

    .. versionadded:: 2015.8.0

    CLI Example:

    .. code-block:: bash

        salt '*' riak.services
    """
    cmd = __execute_cmd("riak-admin", "services")

    return cmd["stdout"][1:-1].split(",")



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.grafana as grafana
import salt.utils.json
from salt.exceptions import SaltInvocationError
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {grafana: {}}


def test_dashboard_present():
    """
    Test to ensure the grafana dashboard exists and is managed.
    """
    name = "myservice"
    rows = ["systemhealth", "requests", "title"]
    row = [{"panels": [{"id": "a"}], "title": "systemhealth"}]

    ret = {"name": name, "result": None, "changes": {}, "comment": ""}

    comt1 = (
        "Dashboard myservice is set to be updated. The following rows "
        "set to be updated: {}".format(["systemhealth"])
    )
    pytest.raises(SaltInvocationError, grafana.dashboard_present, name, profile=False)

    pytest.raises(SaltInvocationError, grafana.dashboard_present, name, True, True)

    mock = MagicMock(
        side_effect=[
            {"hosts": True, "index": False},
            {"hosts": True, "index": True},
            {"hosts": True, "index": True},
            {"hosts": True, "index": True},
            {"hosts": True, "index": True},
            {"hosts": True, "index": True},
            {"hosts": True, "index": True},
        ]
    )
    mock_f = MagicMock(side_effect=[False, False, True, True, True, True])
    mock_t = MagicMock(return_value="")
    mock_i = MagicMock(return_value=False)
    source = {"dashboard": '["rows", {"rows":["baz", null, 1.0, 2]}]'}
    mock_dict = MagicMock(return_value={"_source": source})
    with patch.dict(
        grafana.__salt__,
        {
            "config.option": mock,
            "elasticsearch.exists": mock_f,
            "pillar.get": mock_t,
            "elasticsearch.get": mock_dict,
            "elasticsearch.index": mock_i,
        },
    ):
        pytest.raises(SaltInvocationError, grafana.dashboard_present, name)

        with patch.dict(grafana.__opts__, {"test": True}):
            pytest.raises(SaltInvocationError, grafana.dashboard_present, name)

            comt = f"Dashboard {name} is set to be created."
            ret.update({"comment": comt})
            assert grafana.dashboard_present(name, True) == ret

            mock = MagicMock(
                return_value={"rows": [{"panels": "b", "title": "systemhealth"}]}
            )
            with patch.object(salt.utils.json, "loads", mock):
                ret.update({"comment": comt1, "result": None})
                assert grafana.dashboard_present(name, True, rows=row) == ret

        with patch.object(
            salt.utils.json, "loads", MagicMock(return_value={"rows": {}})
        ):
            pytest.raises(
                SaltInvocationError,
                grafana.dashboard_present,
                name,
                rows_from_pillar=rows,
            )

            comt = "Dashboard myservice is up to date"
            ret.update({"comment": comt, "result": True})
            assert grafana.dashboard_present(name, True) == ret

        mock = MagicMock(
            return_value={"rows": [{"panels": "b", "title": "systemhealth"}]}
        )
        with patch.dict(grafana.__opts__, {"test": False}):
            with patch.object(salt.utils.json, "loads", mock):
                comt = "Failed to update dashboard myservice."
                ret.update({"comment": comt, "result": False})
                assert grafana.dashboard_present(name, True, rows=row) == ret


def test_dashboard_absent():
    """
    Test to ensure the named grafana dashboard is deleted.
    """
    name = "myservice"

    ret = {"name": name, "result": None, "changes": {}, "comment": ""}

    mock = MagicMock(
        side_effect=[
            {"hosts": True, "index": False},
            {"hosts": True, "index": True},
            {"hosts": True, "index": True},
        ]
    )
    mock_f = MagicMock(side_effect=[True, False])
    with patch.dict(
        grafana.__salt__, {"config.option": mock, "elasticsearch.exists": mock_f}
    ):
        pytest.raises(SaltInvocationError, grafana.dashboard_absent, name)

        with patch.dict(grafana.__opts__, {"test": True}):
            comt = "Dashboard myservice is set to be removed."
            ret.update({"comment": comt, "result": None})
            assert grafana.dashboard_absent(name) == ret

        comt = "Dashboard myservice does not exist."
        ret.update({"comment": comt, "result": True})
        assert grafana.dashboard_absent(name) == ret
